/* Intro {{{
 * ----------------------------------------------------------------
 * DConnect Daemon
 *
 *	 #(@) Copyright (c) 2002, DConnect development team
 *	 #(@) Homepage: http://www.dc.ds.pg.gda.pl/
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
 * MA 02139, USA.
 *
 * ----------------------------------------------------------------
 * }}} */

/* include local header files {{{ */
#include "pch.h"
#include "main.h"
#include "conf.h"
#include "net.h"
#include "misc.h"
#include "match.h"
#include "penalties.h"
#include "cmd.user.h"
/* }}} */

extern int NPENAL; /* penalties count */
extern int NUSERS; /* penalties count */

extern Tpenalty *penalties[MAXPENALTIES]; /* table of penalties */
extern userrec_t *user[MAXUSERS];	/* users table */
extern config_t conf;		/* configuration data */

extern int rehash;

extern pthread_mutex_t mutex_myinfo;

/* the assistant function {{{ */

/* cmd_nick2id(*usr,*nick)
.description.
			gets the id for the nick; if the nick is not specified or user does not exist
			apropriate message is shown to the usr

.args.
			userrec_t *usr	-IN-	 sending user's userec
			char *nick				-IN-	 the nick, whose id we want

.return values.
			<value> - the id of the user
			-1	 - user not exists or nick not specified
{{{ */
int cmd_nick2id(userrec_t *usr, char *nick)
{
	int n;

	if (!nick || !*nick)
	{
		pubmsg(usr,"	User nick needed");
		pubmsg(usr,"========================================");
		return -1;
	}
	n=nick2id(nick);

	if (n<0)
	{

		pubmsg(usr,"	User '%s' not found",nick);
		pubmsg(usr,"========================================");
		return -1;
	}
	return n;
} /* }}} */

/* fmthln(*usr, *perm, console, visibility, *commnd, *description)
.description.
		displays formatted help line to >usr< whose usr->perm matches one of the >perm<
		the line format is
		in chat: <HUB>#commmand		 <- decription
		in cons: commmand				<- decription

.args.
		userrec_t *usr - IN -	the user record; if usr is console user, does not display '#' before commnd
		char *perm			- IN - contains permissions i.e:"cas"; only users with one of thease will see the line
		int console		- IN - determines if line is visible from console, chat or both
		char *commnd		- IN - contains the command; if no specified, does not display '#' before commnd
		char *description	- IN - contains description of the command; if no description is specified, display commnd as <HUB> commnd ; else display <HUB> (empty line)

.exaple use.
		fmthln(usr,"cas","Test",NULL);	---displays-->	'<HUB> Test'
		fmthln(usr,"cas",NULL,"Test");	---displays-->	'					Test'
		fmthln(usr,"cas","Axe","Test");	---displays-->	'#Axe				Test'

{{{*/
void fmthln(userrec_t *usr,char *perm, int visibility, char *commnd, char *description)
{
	if ((perm) && (usr->perm) &&(
		!(strchr(usr->perm,'s') && strchr(perm,'s')) &&
		!(strchr(usr->perm,'c') && strchr(perm,'c')) &&
		!(strchr(usr->perm,'a') && strchr(perm,'a')) )) return;

	if (!(usr->cons && visibility&VIS_CONS) && !(!(usr->cons) && visibility&VIS_CHAT)) return;

	//if user is console user or no commnd is specified, display '#' before commnd
	if (description) pubmsg(usr,(commnd)?"%c%-30s<- %s":"%c%-30s %s" ,(usr->cons==0)?((!commnd)?' ':'#'):' ',(commnd)?commnd:"",description);
	//if no description is specified, display commnd as pubmsg; else display empty line as pubmsg
	else pubmsg(usr,"%s",(commnd)?commnd:"");
}/* }}} */

/* penalprvmsg(*to,*from,*fmt,...)
.description.
	sends the private message to from 'from' userrec 'to' userrec;
	the message contains type of penalty and reason;
.args.
	userrec_t *to	//recievers's userrec
	userrec_t *from	//sender's userrec
	char *message		//the message
{{{ */
void penalprvmsg(userrec_t *to, char *op, char *fmt, ...)
{
	char *str=NULL;
	va_list args;

	va_start(args , fmt);

	debug(DEBUG_PROTO,"penalprvmsg() enter");

	str=(char *)my_vsprintf(fmt,args);
	privmsg(to,op,str);
	va_end(args);

	if ((to->penalty) && (to->penalty->reason))
	{
		privmsg(to,op,"R e a s o n:");
		privmsg(to,op,"%s",to->penalty->reason==NULL?"NO REASON":to->penalty->reason);
	}
	my_free(str);
	debug(DEBUG_PROTO,"penalprvmsg() leave");
} /* }}} */

/* penalmsg()
.description.
	sends the penalty messeges (pub, priv and debug)
	kicks banned users;
.args.
	Tpenalty *penrec	//info about penalty

{{{ */
void penalmsg(Tpenalty *penrec)
{
	int i;
	long days, hours, min, sec, time_;

	debug(DEBUG_PROTO,"penalmsg() enter");

	time_=user_exp_penalty(penrec);

	days=time_/(long)(24*60*60);
	time_=time_%(24*60*60);

	hours=time_/(long)(60*60);
	time_=time_%(60*60);
	min=time_/(long)60;
	time_=time_%60;

	sec=time_;

	//show info about punished nick
	pubmsg(NULL,"'%s' has no permission to: %s%s%s%s%s",penrec->nick,
							    (penrec->penalty&PENALTY_BANNED)?"[enter the hub (ban)] ":"",
							    (penrec->penalty&PENALTY_NODL)?"[download] ":"",
							    (penrec->penalty&PENALTY_NOSEARCH)?"[use search] ":"",
							    (penrec->penalty&PENALTY_NOPBLCHAT)?"[chat in public] ":"",
							    (penrec->penalty&PENALTY_NOPRVCHAT)?"[chat in private] ":"");

	pubmsg(NULL,"for %ld days and %ld:%ld:%ld because: %s.",days,
								hours,
								min,
								sec,
								penrec->reason);

	/*add the item to the penalties list add the penalties to userrec*/
	for (i=0;i<NUSERS;i++)
		if (user[i] && user[i]->perm && !strchr(user[i]->perm,'s') && ((penrec==user[i]->penalty) || (user[i]->ip && match(user[i]->ip,penrec->ip))))
		{
			user[i]->penalty=penrec;
			penalprvmsg(user[i],penrec->op,"You were punished as '%s'. You can't %s%s%s%s%s for %ld days and %ld:%ld:%ld.",
													 penrec->nick,
													 (penrec->penalty&PENALTY_BANNED)?"[enter the hub (ban)] ":"",
													 (penrec->penalty&PENALTY_NODL)?"[download] ":"",
													 (penrec->penalty&PENALTY_NOSEARCH)?"[use search] ":"",
													 (penrec->penalty&PENALTY_NOPBLCHAT)?"[chat in public] ":"",
													 (penrec->penalty&PENALTY_NOPRVCHAT)?"[chat in private] ":"",
													 days,
													 hours,
													 min,
													 sec);

			debug(DEBUG_NET,"'%s' punished '%s'@%s by %s%s%s%s%s for %ld days and %ld:%ld:%ld because: %s.",
													 penrec->op,
													 user[i]->nick,
													 penrec->ip,
													 (penrec->penalty&PENALTY_BANNED)?"[ban] ":"",
													 (penrec->penalty&PENALTY_NODL)?"[nodl] ":"",
													 (penrec->penalty&PENALTY_NOSEARCH)?"[nosearch] ":"",
													 (penrec->penalty&PENALTY_NOPBLCHAT)?"[nopblchat] ":"",
													 (penrec->penalty&PENALTY_NOPRVCHAT)?"[noprvchat] ":"",
													 days,
													 hours,
													 min,
													 sec,
													 user[i]->penalty->reason);

			if (penrec->penalty&PENALTY_BANNED) user_set_state(user[i],STATE_QUIT);
		}
	debug(DEBUG_PROTO,"penalmsg() leave");
} /* }}} */

/* opbreak - user with less permissions can not run a command on OP or sOP {{{ */
int opbreak(userrec_t *op,userrec_t *usr, char *commnd)
{
	if(!commnd)
	{
		pubmsg(usr,"ERROR: in 'opbreak': command not specified; returning 1; ");
		return 1;
	}

	if (usr->perm!=NULL && op->perm!=NULL)
	{
		if (op->cons && strchr(op->perm,'s')) return 0;

		if (strchr(op->perm,'a') && strchr(usr->perm,'a'))
		{
			pubmsg(usr,"	You are trying to %s OP", commnd);
			pubmsg(usr,"========================================");
			return 1;
		}

		if (strchr(usr->perm,'s'))
		{
			pubmsg(usr,"	You are trying to %s sOP", commnd);
			pubmsg(usr,"========================================");
			pubmsg(op,"%s is trying to %s you.", usr->nick,commnd);
			return 1;
		}
	}

	return 0;

}/* }}} */

/* }}} */
/* the COMMAND functions {{{ */

/* cmdhelp - shows usr the help screen {{{*/
void cmd_help( userrec_t *usr,char *line)
{
	fmthln(usr,NULL, VIS_ALL,	"========================================",	NULL);
	fmthln(usr,NULL, VIS_ALL, "HELP",	NULL);
	fmthln(usr,NULL, VIS_ALL, "========================================", NULL);

	fmthln(usr,NULL, VIS_ALL,	"[commands listed by categories]",	NULL);
	fmthln(usr,NULL, VIS_ALL, NULL,NULL);

	fmthln(usr,NULL, VIS_ALL,	"[info]:",	NULL);
	fmthln(usr,NULL, VIS_ALL,	"-----------------------------------------", NULL);
	fmthln(usr,"c", VIS_CONS,	"conf",			"shows current hub configuration");
	fmthln(usr,"c",	 VIS_CONS,	"stat",			"shows statistics");
	fmthln(usr,NULL, VIS_ALL,	"rules",		"shows rules which must be respected on the hub");
	fmthln(usr,NULL, VIS_ALL,	"uptime",		"shows hub uptime");
	fmthln(usr,"c", VIS_CONS,	"users",		"shows all hub users");
	fmthln(usr,NULL, VIS_ALL,	"version",		"shows version of the hub");
	fmthln(usr,"c", VIS_CONS,	"who",			"shows who is on the console");
	fmthln(usr,"sac", VIS_ALL,	"whois <nick>",	"shows info about user");
	fmthln(usr,"sac", VIS_ALL,	"whoip <ip>","shows nicks of every user having specified ip");
	fmthln(usr,NULL, VIS_ALL,	"-----------------------------------------", NULL);
	fmthln(usr,"as", VIS_ALL, NULL,NULL);

	fmthln(usr,"as", VIS_ALL,	"[punishments]:", NULL);
	fmthln(usr,"as", VIS_ALL,	"-----------------------------------------",NULL);
	fmthln(usr,"as", VIS_ALL,	"amnesty <ip>",					"removes penalty from ip");
	fmthln(usr,"as", VIS_ALL,	"amnesty_all",					"clears the punished ip's list");
	fmthln(usr,"as", VIS_ALL,	"lspunished",					"shows the list of punished ip's");
	fmthln(usr,"as", VIS_ALL,	"punish <nick/IP> <how> [how long in minutes] [reason]","punishes the user");
	fmthln(usr,"as", VIS_ALL,	NULL,							"<how>=(nosearch,nopblchat,noprvchat,nodl,ban)");
	fmthln(usr,"as", VIS_ALL,	"kick <nick> <reason>",		"disconnects the user");
	fmthln(usr,"as", VIS_ALL,	"-----------------------------------------",NULL);
	fmthln(usr,"as", VIS_ALL, NULL,NULL);

	fmthln(usr,NULL, VIS_ALL,	"[communication & redir]:" ,NULL);
	fmthln(usr,NULL, VIS_ALL,	"-----------------------------------------" ,NULL);
	fmthln(usr,"as", VIS_ALL,	"msg <nick> <message>"		,"sends message to user in pub chat as hub");
	fmthln(usr,"s", VIS_ALL,	"msg_prv <nick> <message>"	,"sends private message to user in pub chat as hub");
	fmthln(usr,NULL, VIS_ALL,	"mmsg <nick> <message>"		,"sends message to user in pub chat as your message");
	fmthln(usr,"s", VIS_ALL,	"msg_all <message>"			,"sends message to all users in pub chat as hub");
	fmthln(usr,"s", VIS_ALL,	"msg_prv_all <message>"		,"sends private message to all users in pub chat as hub");
	fmthln(usr,NULL, VIS_ALL,	"mmsg_all <message>"		,"sends message to all users in pub chat as your message");
	fmthln(usr,NULL, VIS_CONS,	"pubchat"					,"toggles public chat mode");

	fmthln(usr,"as", VIS_ALL,	"redirect <nick> <where>"	,"moves user to another hub");
	fmthln(usr,"s", VIS_ALL,	"redirect_all <where>"		,"moves all users to another hub");
	fmthln(usr,NULL, VIS_ALL, "-----------------------------------------" ,NULL);
	fmthln(usr,"c", VIS_ALL, NULL ,NULL);

	fmthln(usr,"c", VIS_CONS,	"[other commands]:" ,NULL);
	fmthln(usr,"c", VIS_CONS,	"-----------------------------------------" ,NULL);
	fmthln(usr,"c", VIS_CONS,	"rehash"	,"reloads hub configuration");
	fmthln(usr,"c", VIS_CONS,	"quit"		,"leave the console");
	fmthln(usr,"c", VIS_CONS,	"jq"		,"toggles displaing of joins & quits traffic");
	fmthln(usr,"c", VIS_CONS,	"boost"		,"toggles info-booster");
	fmthln(usr,"c", VIS_CONS,	"-----------------------------------------" ,NULL);
	fmthln(usr,NULL, VIS_ALL, "========================================", NULL);
	return;

} /* }}} */

/* hlp_redirect - shows help for redirect command {{{*/
void hlp_redirect( userrec_t *usr)
{
	pubmsg(usr," ");
	pubmsg(usr,"Description: ");
	pubmsg(usr,"	Moves the user to another hub.");
	pubmsg(usr," ");
	pubmsg(usr,"Syntax:");
	pubmsg(usr,"	%credirect <nick> <where>",(usr->cons)?' ':'#');
	pubmsg(usr," ");
	pubmsg(usr,"	<nick> is the nick of user");
	pubmsg(usr,"	<where> is the adress of hub into <user> should be moved");
	pubmsg(usr,"========================================");
} /* }}} */

/* cmdredirect - redirect's the user to another hub {{{ */
void cmd_redirect( userrec_t *usr,char *line)
{
	int n;
	char *tmp1=NULL, *tmp2=NULL;
	char *__strtok_temp__=NULL;

	//checking if user has apriopriate permissions
	if (usr->perm && !strchr(usr->perm,'s') && !strchr(usr->perm,'a')) return;

	if (!line || !*line) { hlp_redirect( usr ); return;}

	tmp1=strtok_r(line," ",&__strtok_temp__); //who
	tmp2=strtok_r(NULL,"|",&__strtok_temp__);//where

	pubmsg(usr,"========================================");
	pubmsg(usr," REDIRECT");
	pubmsg(usr,"========================================");

	// checks the nick, returns id
	n=cmd_nick2id(usr,tmp1); if(n<0)return;

	if (!tmp2)
	{
		pubmsg(usr,"	Do not know where to redirect");
		pubmsg(usr,"========================================");
		return;
	}

	if (!strcmp(usr->nick,tmp1))
	{
		pubmsg(usr,"	You are trying to move yourself");
		pubmsg(usr,"========================================");
		return;
	}

	if (opbreak(usr,user[n],"redirect"))return;

	debug(DEBUG_NET,"'%s'(OP) moved %s to '%s'",usr->nick,user[n]->nick,tmp2);

	sendf(user[n],"$ForceMove %s|",tmp2);
	user_set_state(user[n],STATE_QUIT);
	pubmsg(usr,"	Redirecting '%s' to %s",tmp1,tmp2);
	pubmsg(usr,"========================================");

	if((n=nick2id(tmp1))>=0) user_set_state(user[n],STATE_QUIT);
} /* }}} */

/* hlp_redirect - shows help for redirect command {{{*/
void hlp_redirect_all( userrec_t *usr)
{
	pubmsg(usr," ");
	pubmsg(usr,"Description: ");
	pubmsg(usr,"	Moves all the users to another hub.");
	pubmsg(usr," ");
	pubmsg(usr,"Syntax:");
	pubmsg(usr,"	%credirect_all <where>",(usr->cons)?' ':'#');
	pubmsg(usr," ");
	pubmsg(usr,"	<where> is the adress of hub into users should be moved.");
	pubmsg(usr,"========================================");
} /* }}} */

/* cmdredirectall - redirect all users to anther hub {{{*/
void cmd_redirect_all( userrec_t *usr,char *line)
{
	int temp=0,n;
	char *tmp1=NULL,*__strtok_temp__;

	//checking if user has apriopriate permissions
	if (usr->perm && !strchr(usr->perm,'s')) return;

	if (!line || !*line) {hlp_redirect_all( usr ); return;}
	tmp1=strtok_r(line,"| ",&__strtok_temp__); //<where>

	pubmsg(usr,"========================================");
	pubmsg(usr," REDIRECT_ALL");
	pubmsg(usr,"========================================");

	if (!tmp1)
	{
		pubmsg(usr,"	do not know where to redirect to");
		pubmsg(usr,"========================================");
		return;
	}

	debug(DEBUG_NET,"'%s'(sOP) moved ALL to '%s'",usr->nick,tmp1);

	pubmsg(usr,"	Moving users to %s",tmp1);

	temp=0;
	for(n=0;n<NUSERS;n++)
		if (user[n] && user_tst_state(user[n],STATE_LOGGEDIN|STATE_SR|STATE_REGISTERED) && usr->perm && !strchr(user[n]->perm,'s'))
		{
			temp++;
			sendf(user[n],"$ForceMove %s|",tmp1);
			user_set_state(user[n],STATE_QUIT);
		}

	pubmsg(usr,"	Communicate sent to %d of %d users",temp,NUSERS);
	pubmsg(usr,"========================================");
	debug(DEBUG_NET,"'%s'(OP) moved ALL to '%s'",usr->nick,tmp1);
} /* }}} */

/* cmdpassword {{{*/
void cmd_password( userrec_t *usr,char *line)
{
	pubmsg(usr,"========================================");
	pubmsg(usr," PASSWORD");
	pubmsg(usr,"========================================");
	pubmsg(usr,"	Command not supported yet");
	pubmsg(usr,"========================================");
} /* }}} */

/* cmduptime - displays hub uptime {{{ */
void cmd_uptime( userrec_t *usr)
{
	char str[32];
	pubmsg(usr,"========================================");
	pubmsg(usr," UPTIME");
	pubmsg(usr,"========================================");
	uptime(starttime,str);
	pubmsg(usr,"	%s",str);
	pubmsg(usr,"========================================");
} /* }}} */

/* cmdrules - displays dcd.rules {{{ */
void cmd_rules( userrec_t *usr)
{
	pubmsg(usr,"========================================");
	pubmsg(usr," RULES");
	pubmsg(usr,"========================================");
	if (conf.conf_rules) playfile(usr,conf.conf_rules);
	else pubmsg(usr,"No rules specified");
	pubmsg(usr,"========================================");
} /* }}} */


/* cmdversion - displays hub version {{{ */
void cmd_version( userrec_t *usr)
{
	pubmsg(usr,"========================================");
	pubmsg(usr," VERSION");
	pubmsg(usr,"========================================");
	pubmsg(usr,"	%s (build: %s)",PACKAGE_VERSION,BUILD_TIME);
	pubmsg(usr,"========================================");
} /* }}} */

	/* cmdlspunished - displays the list of punished {{{*/
void cmd_lspunished( userrec_t *usr)
{
	int n;
	long time_, days, hours, min, sec;
	//checking if user has apriopriate permissions
	if ( usr->perm && !strchr(usr->perm,'s') && !strchr(usr->perm,'a')) return;

	pubmsg(usr,"========================================");
	pubmsg(usr," LSPUNISHED");
	pubmsg(usr,"========================================");

	if (NPENAL==0) pubmsg(usr,"	Nobody is punished :(");
	else pubmsg(usr,"	IP : Nick: Punisher : Penalty : Time(d,hh:mm:ss) : Reason");

	for(n=0;n<NPENAL;n++)
	{

		time_=user_exp_penalty(penalties[n]);
		if (time_<=0) continue;

		days=time_/(long)(24*60*60);
		time_=time_%(24*60*60);

		hours=time_/(long)(60*60);
		time_=time_%(60*60);

		min=time_/(long)60;
		time_=time_%60;

		sec=time_;

		pubmsg(usr,"	%s : %s : %s : %s%s%s%s%s : %ld, %ld:%ld:%ld : %s", penalties[n]->ip,
																penalties[n]->nick,
																strchr(usr->perm,'s')?penalties[n]->op:"HUB",
																(penalties[n]->penalty&PENALTY_BANNED)?"[ban] ":"",
																(penalties[n]->penalty&PENALTY_NODL)?"[nodl] ":"",
																(penalties[n]->penalty&PENALTY_NOSEARCH)?"[nosearch] ":"",
																(penalties[n]->penalty&PENALTY_NOPBLCHAT)?"[nopblchat] ":"",
																(penalties[n]->penalty&PENALTY_NOPRVCHAT)?"[noprvchat] ":"",
																days,
																hours,
																min,
																sec,
																penalties[n]->reason);
	}
	pubmsg(usr,"========================================");
} /* }}} */

/* hlp_punish - displays help for punish command {{{*/
void hlp_punish( userrec_t *usr)
{
	pubmsg(usr,"	Punishment not specified");
	pubmsg(usr,"	Available punishments sample use:");
	pubmsg(usr," ");
	pubmsg(usr,"	%cpunish <nick/ip> <how> [time] [reason]",(usr->cons)?' ':'#');
	pubmsg(usr," ");
	pubmsg(usr,"	<how>");
	pubmsg(usr,"	nosearch		- user has no permission to use search");
	pubmsg(usr,"	nopblchat		- user has no permission to chat in main chat");
	pubmsg(usr,"	noprvchat		- user has no permission to chat in private chat");
	pubmsg(usr,"	nochat			- user has no permission to	chan in main and private chat");
	pubmsg(usr,"	nodl			- user has no permission to download");
	pubmsg(usr,"	ban				- user is banned");
	pubmsg(usr," ");
	pubmsg(usr,"	You can punish with multiple punishments, merging them with '+'");
	pubmsg(usr,"	i.e. ");
	pubmsg(usr,"		%cpunish UserNick nochat+nodl+nosearch 20",(usr->cons)?' ':'#');
	pubmsg(usr," ");
	pubmsg(usr,"	[time]			- the duration of punishment in minutes");
	pubmsg(usr," ");
	pubmsg(usr,"	[time] and [reason] are optional");
	pubmsg(usr,"	if no time is specified the duration of penalty will be %ld minute(s)",STD_PENALTY_DURATION);
	pubmsg(usr," ");
	pubmsg(usr,"	1 day - 1440 minutes");
	pubmsg(usr,"	3 days - 4320 minutes");
	pubmsg(usr,"	1 week - 10080 minutes");
	pubmsg(usr,"	eternity (1 year) - -1 minutes :)");
	pubmsg(usr," ");
	pubmsg(usr,"	examples:");
	pubmsg(usr,"		%cpunish UserNick nochat 10 Was naughty.",(usr->cons)?' ':'#');
	pubmsg(usr,"		%cpunish UserNick nochat+nodl",(usr->cons)?' ':'#');
	pubmsg(usr,"		%cpunish UserNick nosearch Was naughty.",(usr->cons)?' ':'#');
	pubmsg(usr,"		%cpunish UserIP nochat 10 Was naughty.",(usr->cons)?' ':'#');
	pubmsg(usr,"========================================");
} /* }}} */

/* cmdpunish - punishes the user with one of the avaliable penalties {{{ */
void cmd_punish ( userrec_t *usr,char *line)
{
	const char *punish[]=
	{
		"nopblchat",
		"ban",
		"nosearch",
		"noprvchat",
		"nodl",
		"nochat",
		"" /* indicates end of list */
	};
//	long days, hours, min, sec, time;
	long /*i=0,*/ n=0,countpunish=0, duration=0;
	char *who_tmp=NULL,*how_tmp=NULL,*msg_tmp=NULL,*tmp=NULL,*tmp1=NULL,*pun=NULL, *__strtok_temp__;
	int punishment=0;
 	int a; /* for checking ip */
 	
	/*checking if user has apriopriate permissions*/
	if ( usr->perm && !strchr(usr->perm,'s') && !strchr(usr->perm,'a'))return;
	pubmsg(usr,"========================================");
	pubmsg(usr," PUNISH");
	pubmsg(usr,"========================================");

	if (!line || !*line)
	{
		hlp_punish(usr);
		return;
	}

	who_tmp=strtok_r(line," ",&__strtok_temp__); //who
	how_tmp=strtok_r(NULL," ",&__strtok_temp__); //how
	tmp=strtok_r(NULL,"",&__strtok_temp__); //time + message

	if (tmp)
	{
		duration=strtol(tmp,&tmp1,10);

		if (duration==0) duration=conf.std_penalty_duration;
		if (*tmp1) msg_tmp=tmp1;
	}
	else duration=conf.std_penalty_duration;

	if (!who_tmp || !how_tmp)
	{
		hlp_punish(usr) ;
		return;
	}

	if (duration==-1) duration=365*24*60; //-1 is one year

	duration=duration*60;// to get duration in seconds

	pun=strtok_r(how_tmp,"+",&__strtok_temp__);

	while(pun && *pun)
	{
		/* parse punishment {{{ */
		countpunish=-1;
		while(*punish[++countpunish]) // was ...nish++], and didn't work
			if (pun && !strcasecmp(pun,punish[countpunish])) break;
		how_tmp=pun;								/* tmp2 contains current punishment(s) */
		pun=strtok_r(NULL,"+",&__strtok_temp__);

		switch(countpunish)
		{
			case 0: //nopblchat
					punishment|=(int)PENALTY_NOPBLCHAT;
					continue;

			case 1: //ban
					punishment|=(int)PENALTY_BANNED;
					continue;

			case 2: //nosearch
					punishment|=(int)PENALTY_NOSEARCH;
					continue;

			case 3: //noprvchat
					punishment|=(int)PENALTY_NOPRVCHAT;
					continue;

			case 4: //nodl
					punishment|=(int)PENALTY_NODL;
					continue;

			case 5: //nochat
					punishment|=(int)PENALTY_NOPRVCHAT|PENALTY_NOPBLCHAT;
					continue;

			default:
					pubmsg(usr,"Unknown punishment '%s'. Type '#punish' for more help",how_tmp);
					break;
		}
/* end of parse punishment }}}*/
	}

	if (punishment)
	{
		n=cmd_nick2id(usr,who_tmp);
		if(n<0)
		{
			if (sscanf(who_tmp,"%d.%d.%d.%d",&a,&a,&a,&a)!=4) return;
			user_set_penalty(who_tmp,"(offline)",usr->nick, punishment,(unsigned long)time(NULL),duration,msg_tmp,0);	
		}
		else
		{
			if ( user[n]->perm && strchr(user[n]->perm,'s')) return;
			user_set_penalty(user[n]->ip,user[n]->nick,usr->nick, punishment,(unsigned long)time(NULL),duration,msg_tmp,0);
		}
	}

	pubmsg(usr,"========================================");

} /* }}} */

/* hlp_msg - shows help for msg command {{{*/
void hlp_msg( userrec_t *usr)
{
	pubmsg(usr,"========================================");
	pubmsg(usr,"	MSG");
	pubmsg(usr,"========================================");
	pubmsg(usr," ");
	pubmsg(usr,"Description: ");
	pubmsg(usr,"	Shows the <message> to the <user> as 'Hub'");
	pubmsg(usr," ");
	pubmsg(usr,"Syntax:");
	pubmsg(usr,"	%cmsg <nick> <message>",(usr->cons)?' ':'#');
	pubmsg(usr," ");
	pubmsg(usr,"========================================");
} /* }}} */

/* cmd_msg - sends message as hub to the user {{{*/
void cmd_msg( userrec_t *usr,char *line)
{
	int n;
	char *tmp1=NULL,*tmp2=NULL;
	char *__strtok_temp__=NULL;

	//checking if user has apriopriate permissions
	if ( usr->perm && !strchr(usr->perm,'s') && !strchr(usr->perm,'a')) return;
	if (!line || !*line) {hlp_msg(usr); return;}
	tmp1=strtok_r(line," ",&__strtok_temp__); //who
	tmp2=strtok_r(NULL,"",&__strtok_temp__); //message

	n=cmd_nick2id(usr,tmp1); 
	if(n<0)return;

	if (!tmp2)
	{
		pubmsg(usr,"========================================");
		pubmsg(usr,"	MSG");
		pubmsg(usr,"========================================");
		pubmsg(usr,"	You have to specify message");
		pubmsg(usr,"========================================");
		return;
	}
	if (user[n]!=usr) pubmsg(user[n],tmp2);
}/* }}} */

/* hlp_msg_prv - shows help for msg_prv command {{{*/
void hlp_msg_prv( userrec_t *usr)
{
	pubmsg(usr,"========================================");
	pubmsg(usr,"	MSG");
	pubmsg(usr,"========================================");
	pubmsg(usr," ");
	pubmsg(usr,"Description: ");
	pubmsg(usr,"	Sends private <message> to the <user> as 'Hub'");
	pubmsg(usr," ");
	pubmsg(usr,"Syntax:");
	pubmsg(usr,"	%cmsg <nick> <message>",(usr->cons)?' ':'#');
	pubmsg(usr," ");
	pubmsg(usr,"========================================");
} /* }}} */

/* cmd_msg_prv - sends message as hub to the user {{{*/
void cmd_msg_prv( userrec_t *usr,char *line)
{
	int n;
	char *tmp1=NULL,*tmp2=NULL;
	char *__strtok_temp__=NULL;

	//checking if user has apriopriate permissions
	if ( usr->perm && !strchr(usr->perm,'s')) return;
	if (!line || !*line) { hlp_msg_prv(usr); return;}
	tmp1=strtok_r(line," ",&__strtok_temp__); //who
	tmp2=strtok_r(NULL,"",&__strtok_temp__); //message

	n=cmd_nick2id(usr,tmp1); 
	if(n<0)return;

	if (!tmp2)
	{
		pubmsg(usr,"========================================");
		pubmsg(usr,"	MSG");
		pubmsg(usr,"========================================");
		pubmsg(usr,"	You have to specify message");
		pubmsg(usr,"========================================");
		return;
	}

	if (user[n]!=usr) privmsg(user[n],NULL,tmp2);
}/* }}} */

/* hlp_msg_prv_all - shows help for msg_prv_all command {{{*/
void hlp_msg_prv_all( userrec_t *usr)
{
	pubmsg(usr,"========================================");
	pubmsg(usr,"	MSG");
	pubmsg(usr,"========================================");
	pubmsg(usr," ");
	pubmsg(usr,"Description: ");
	pubmsg(usr,"	Sends private <message> to all the users as 'Hub'");
	pubmsg(usr," ");
	pubmsg(usr,"Syntax:");
	pubmsg(usr,"	%cmsg_prv_all <message>",(usr->cons)?' ':'#');
	pubmsg(usr," ");
	pubmsg(usr,"========================================");
} /* }}} */

/* cmdmsg - sends private message as hub to the user {{{*/
void cmd_msg_prv_all( userrec_t *usr,char *line)
{
	int n;
	char *tmp1=NULL,*tmp2=NULL;
	char *__strtok_temp__=NULL;

	//checking if user has apriopriate permissions
	if ( usr->perm && !strchr(usr->perm,'s')) return;
	if (!line || !*line) { hlp_msg_prv_all(usr); return;}

	tmp1=strtok_r(line,"",&__strtok_temp__); //who


	n=cmd_nick2id(usr,tmp1); 
	if(n<0)return;

	if (!tmp1)
	{
		pubmsg(usr,"========================================");
		pubmsg(usr,"	MSG");
		pubmsg(usr,"========================================");
		pubmsg(usr,"	You have to specify message");
		pubmsg(usr,"========================================");
		return;
	}

	if (user[n]!=usr) 		privmsg(NULL,NULL,tmp2);
}/* }}} */



/* hlp_msg - shows help for msg command {{{*/
void hlp_mmsg( userrec_t *usr)
{
	pubmsg(usr,"========================================");
	pubmsg(usr,"	MMSG");
	pubmsg(usr,"========================================");
	pubmsg(usr," ");
	pubmsg(usr,"Description: ");
	pubmsg(usr,"	Shows the <message> to the <user> as Your message");
	pubmsg(usr," ");
	pubmsg(usr,"Syntax:");
	pubmsg(usr,"	%cmmsg <nick> <message>",(usr->cons)?' ':'#');
	pubmsg(usr," ");
	pubmsg(usr,"========================================");
} /* }}} */

/* cmdmsg - sends message as hub to the user {{{*/
void cmd_mmsg( userrec_t *usr,char *line)
{
	int n;
	char *tmp1=NULL,*tmp2=NULL;
	char *__strtok_temp__=NULL;

	//checking if user has apriopriate permissions
	if (!line || !*line) {hlp_msg(usr); return;}
	tmp1=strtok_r(line," ",&__strtok_temp__); //who
	tmp2=strtok_r(NULL,"",&__strtok_temp__); //message

	n=cmd_nick2id(usr,tmp1); if(n<0)return;

	if (!tmp2)
	{
		pubmsg(usr,"========================================");
		pubmsg(usr,"	MSG");
		pubmsg(usr,"========================================");
		pubmsg(usr,"	You have to specify message");
		pubmsg(usr,"========================================");
		return;
	}
	if (user[n]!=usr) sendf(user[n],"<%s> %s|",usr->nick,tmp2);
}/* }}} */

/* hlp_pubchat - sends message as hub to the user {{{*/
void hlp_pubchat( userrec_t *usr)
{
	pubmsg(usr,"========================================");
	pubmsg(usr,"	PUBCHAT");
	pubmsg(usr,"========================================");
	pubmsg(usr," ");
	pubmsg(usr,"Description: ");
	pubmsg(usr,"	enables for a console user chating freely");
	pubmsg(usr," ");
	pubmsg(usr,"Syntax:");
	pubmsg(usr,"	%cpubchat",(usr->cons)?' ':'#');
	pubmsg(usr," ");
	pubmsg(usr,"========================================");
} /* }}} */

/* cmd_mmsg - sends message as usr to the user {{{*/
void cmd_pubchat( userrec_t *usr,char *line)
{
	//checking if user has apriopriate permissions
	if (!usr->cons) return;

	if (usr->cons&2)
	{
		pubmsg(usr,"Chating in public for you is now off");
		usr->cons=1;
	}
	else
		{
			usr->cons=3;
			pubmsg(usr,"Chating in public for you is now on");
			pubmsg(usr,"To stop. write 'pubchat' one more time");
		}
}/* }}} */

/* hlp_msg_all - shows help for msg command {{{*/
void hlp_msg_all( userrec_t *usr)
{
	pubmsg(usr,"========================================");
	pubmsg(usr,"	MSG_ALL");
	pubmsg(usr,"========================================");
	pubmsg(usr," ");
	pubmsg(usr,"Description: ");
	pubmsg(usr,"	Shows the <message> to all users as 'Hub'");
	pubmsg(usr," ");
	pubmsg(usr,"Syntax:");
	pubmsg(usr,"	%cmsg_all <message>",(usr->cons)?' ':'#');
	pubmsg(usr," ");
	pubmsg(usr,"========================================");
} /* }}} */

/* cmdmsgall - sends message as hub to all users {{{*/
void cmd_msg_all( userrec_t *usr,char *line)
{
	//checking if user has apriopriate permissions
	if ( usr->perm && !strchr(usr->perm,'s')) return;
	if (!line|| !*line) {hlp_msg_all(usr); return;}

	pubmsg(NULL,line);
	debug(DEBUG_CHAT,"<HUB>(%s) -> ALL: %s", usr->nick,line);
	//DEBUGMSG
} /* }}} */

/* hlp_msg_all - shows help for msg command {{{*/
void hlp_mmsg_all( userrec_t *usr)
{
	pubmsg(usr,"========================================");
	pubmsg(usr,"	MMSG_ALL");
	pubmsg(usr,"========================================");
	pubmsg(usr," ");
	pubmsg(usr,"Description: ");
	pubmsg(usr,"	Shows the <message> to all users as your message");
	pubmsg(usr," ");
	pubmsg(usr,"Syntax:");
	pubmsg(usr,"	%cmmsg_all <message>",(usr->cons)?' ':'#');
	pubmsg(usr," ");
	pubmsg(usr,"========================================");
} /* }}} */

/* cmdmsgall - sends message as hub to all users {{{*/
void cmd_mmsg_all( userrec_t *usr,char *line)
{
	//checking if user has apriopriate permissions
	if (!line|| !*line) {hlp_msg_all(usr); return;}

	sendf(NULL,"<%s> %s|",usr->nick,line);

	//DEBUGMSG
} /* }}} */

/* cmd_amnesty - cancels all of current user's penalties {{{ */
void cmd_amnesty ( userrec_t *usr,char *line)
{
	char *tmp1=NULL;
	char *__strtok_temp__=NULL;
	//checking if user has apriopriate permissions
	if ( usr->perm && !strchr(usr->perm,'s') && !strchr(usr->perm,'a')) return;
	pubmsg(usr,"========================================");
	pubmsg(usr,"	AMNESTY");
	pubmsg(usr,"========================================");
	if (!line|| !*line) return;
	tmp1=strtok_r(line," ",&__strtok_temp__); //ip

	if (!tmp1)
	{
		pubmsg(usr,"	User's nick or ip needed");
		pubmsg(usr,"========================================");
		return;
	}

	debug(DEBUG_NET,"Amnesty for '%s' by %sOP '%s'",tmp1,strchr(usr->perm,'s')?"s":"",usr->nick);

	user_set_penalty(tmp1,tmp1,"",PENALTY_NONE,0,0,NULL,0);

	pubmsg(usr,"	Amnesty of USER/IP: %s",tmp1);
	pubmsg(usr,"========================================");
	//DEBUGMSG
} /* }}} */

 /* cmdamnestyall - cancels all of penalties {{{ */
void cmd_amnesty_all ( userrec_t *usr)
{
	int i,max=NPENAL;

	//checking if user has apriopriate permissions
	if ( usr->perm && !strchr(usr->perm,'s')) return;

	pubmsg(usr,"========================================");
	pubmsg(usr,"	AMNESTY_ALL");
	pubmsg(usr,"========================================");

	debug(DEBUG_NET,"Amnesty for ALL by %sOP '%s'",usr->nick,strchr(usr->perm,'s')?"s":"");

	for(i=0; i<max;i++) delpenalty(0);
	penalties_write();

	pubmsg(usr,"	Amnesty of all punished");
	pubmsg(usr,"========================================");
	//DEBUGMSG
} /* }}} */

/* cmd_conf {{{ */
void cmd_conf ( userrec_t *usr)
{
	//can be used only from console
	if (!usr->cons) return;

	pubmsg(usr,"========================================");
	pubmsg(usr,"	CONF");
	pubmsg(usr,"========================================");
	pubmsg(usr,"Config file: %s", conf.conf_main);
	pubmsg(usr,"Welcome file: %s", conf.conf_welcome);
	pubmsg(usr,"MOTD file: %s", conf.conf_motd);
	pubmsg(usr,"Rules file: %s", conf.conf_rules);
	pubmsg(usr,"Bans file: %s", conf.conf_banned);
	pubmsg(usr,"Console users file: %s", conf.conf_cusers);
	pubmsg(usr,"Console allow file: %s", conf.conf_callow);
	pubmsg(usr,"Nicks allow file: %s", conf.conf_nallow);
	pubmsg(usr,"========================================");
	pubmsg(usr,"Log level: %u", conf.log_level);
	pubmsg(usr,"========================================");
	pubmsg(usr,"Log files dir: %s", conf.log_dir);
	pubmsg(usr,"Main log: %s", conf.log_main);
	pubmsg(usr,"Network access log: %s", conf.log_net);
	pubmsg(usr,"Protocol log: %s", conf.log_proto);
	pubmsg(usr,"Warnings log: %s", conf.log_wrng);
	pubmsg(usr,"Console log: %s", conf.log_con);
	pubmsg(usr,"Public chat log: %s", conf.log_chat);
	pubmsg(usr,"========================================");
	pubmsg(usr,"Main listening port[TCP]: %u", conf.listen_port_main);

	if (conf.listen_port_cons) pubmsg(usr,"Console listening port[TCP]: %u", conf.listen_port_cons);
	if (conf.minslots) pubmsg(usr,"Minslots checking port[UDP]: %u", conf.listen_port_sr);

	pubmsg(usr,"Main Listening threads: %u", conf.n_listen_main);
	pubmsg(usr,"User managing threads: %u", conf.n_user_manager);
	pubmsg(usr,"========================================");
	pubmsg(usr,"UID: %s(%u)", conf.user, conf.uid);
	pubmsg(usr,"GID: %s(%u)", conf.group, conf.gid);
	pubmsg(usr,"========================================");
	pubmsg(usr,"Hub Name: %s", conf.hubname);
	pubmsg(usr,"Admin contact: %s", conf.admin_contact);
	pubmsg(usr,"User limit: %u", conf.userlimit);
	pubmsg(usr,"Allow broken keys: %d", conf.allow_broken_key);
	pubmsg(usr,"Idle timeout: %u", conf.idle_timeout);
	pubmsg(usr,"Registration timeout: %u", conf.register_timeout);
	pubmsg(usr,"penalties update interval (s): %u", conf.penalties_update_interval);
	pubmsg(usr,"Search interval (s): %u", conf.search_interval);
	pubmsg(usr,"Minimum data shared: %"PRIu64" MB", conf.minshare);
	pubmsg(usr,"Minimum slots: %u", conf.minslots);
	pubmsg(usr,"Max commands: %u", conf.max_commands);	
	pubmsg(usr,"Allow chat: %d", conf.allow_chat);
	pubmsg(usr,"Flags: %u", conf.flags);
	pubmsg(usr,"========================================");

	if (conf.redirect_switch)
	{
		pubmsg(usr,"Redirections:");
		if(conf.redirect_switch&REDIRECT_SWITCH_ACCESS)	 pubmsg(usr,"Access:	 %s",conf.redirect_access);
		if(conf.redirect_switch&REDIRECT_SWITCH_HUB_IS_FULL) pubmsg(usr,"Hub is full: %s",conf.redirect_hub_is_full);
		if(conf.redirect_switch&REDIRECT_SWITCH_MINSHARE)	pubmsg(usr,"Min_share: %s",conf.redirect_minshare);
		if(conf.redirect_switch&REDIRECT_SWITCH_MINSLOTS)	pubmsg(usr,"Min_slots: %s",conf.redirect_minslots);
		pubmsg(usr,"========================================");
	}

} /* }}} */

/* cmd_rehash {{{ */
void cmd_rehash( userrec_t *usr)
{
	//can be used only from console
	if (!usr->cons) return;

	pubmsg(usr,"========================================");
	pubmsg(usr,"	REHASH");
	pubmsg(usr,"========================================");
	
	rehash=1;
//	finish(SIGHUP);
} /* }}} */

/* cmd_stat {{{ */
void cmd_stat ( userrec_t *usr)
{
	uint64_t tmp=0;
	int i, reg=0;
	char str[32];

	//can be used only from console
	if (!usr->cons) return;

	for(i=0;i<NUSERS;i++)
		if (user_tst_state(user[i],STATE_REGISTERED))
		{
			tmp+=user[i]->myinfo.share;
			reg++;
		}

	tmp=tmp/(int)(1024*1024);

	i=0; // MB
	if (tmp>=1024)
	{
		tmp=tmp/(int)1024;
		i=1; // GB

		if (tmp>=1024)
		{
			tmp=tmp/(int)1024;
			i=2; // TB
		}
	}

	pubmsg(usr,"========================================");
	pubmsg(usr,"	STAT");
	pubmsg(usr,"========================================");
	pubmsg(usr,"	Users registered: %d",reg);
	pubmsg(usr,"	Users waiting:	%d",NUSERS-reg);
	pubmsg(usr,"	Shared data: %"PRIu64"%cB",tmp,(i==0)?'M':((i==1)?'G':'T'));
	uptime(starttime,str);
	pubmsg(usr,"	Uptime: %s",str);
	pubmsg(usr,"========================================");
} /* }}} */

/* cmd_whois {{{ */
void cmd_whois ( userrec_t *usr,char *line)
{
	int i,count;
	uint64_t tmp;
	char str[32];

	//checking if user has apriopriate permissions
	if (usr->perm && !strchr(usr->perm,'c') &&
		!strchr(usr->perm,'s') &&
		!strchr(usr->perm,'a')) return;

	i=cmd_nick2id(usr,line);

	if (i<0) return;
	
	pthread_mutex_lock(&mutex_myinfo);

	pubmsg(usr,"========================================");
	pubmsg(usr,"	WHOIS");
	pubmsg(usr,"========================================");
	pubmsg(usr,"Nick: %s",user[i]->nick);
	pubmsg(usr,"IP: %s",user[i]->ip);
	pubmsg(usr,"Mode: %s",(user[i]->myinfo.flag==1)?"online":((user[i]->myinfo.flag==3)?"away":"unknown"));
	uptime(user[i]->idle,str);
	pubmsg(usr,"Idle: %s",str);

	if (user[i]->cons==0)
	{
		tmp=user[i]->myinfo.share;
		tmp/=1024*1024; count=0; // MB
		if (tmp>=1024)
		{
			tmp/=1024; count=1; // GB
			if (tmp>=1024) { tmp/=1024; count=2; } // TB
		}
		pubmsg(usr,"Shared: %"PRIu64"%cB",tmp,(count==0)?'M':((count==1)?'G':'T'));
	}

	pubmsg(usr,"E-mail: %s",user[i]->myinfo.email);
	pubmsg(usr,"Connection: %s",user[i]->myinfo.ctype);
	pubmsg(usr,"Description: %s",user[i]->myinfo.desc);
	pubmsg(usr,"Permissions: %s", user[i]->perm?user[i]->perm:"not set");

	if (user[i]->penalty) pubmsg(usr,"Punishments: %d",user[i]->penalty->penalty);
	pubmsg(usr,"========================================");

	pthread_mutex_unlock(&mutex_myinfo);
} /* }}} */

/* cmd_whois {{{ */
void cmd_whoip ( userrec_t *usr,char *line)
{
	int i,count=0;

	//checking if user has apriopriate permissions
	if (usr->perm && !strchr(usr->perm,'c') &&
		!strchr(usr->perm,'s') &&
		!strchr(usr->perm,'a')) return;

	pubmsg(usr,"========================================");
	pubmsg(usr,"	WHOIP");
	pubmsg(usr,"========================================");

	if (!line)
	{
		pubmsg(usr,"You have to specify user's ip");
		pubmsg(usr,"========================================");
		return;
	}

	for(i=0;i<NUSERS;i++)
		if (user[i] &&
			user_tst_state(user[i],STATE_REGISTERED) &&
			!strcmp(user[i]->ip,line))
			{
				pubmsg(usr,"%s",user[i]->nick);
				count++;
			}

	if (!count) pubmsg(usr,"%s","Nobody of this ip was found");
	pubmsg(usr,"========================================");
} /* }}} */

/* cmd_boost {{{ */
void cmd_boost ( userrec_t *usr)
{
	//can be used only from console
	if (!usr->cons) return;

	pubmsg(usr,"========================================");
	pubmsg(usr,"	BOOST");
	pubmsg(usr,"========================================");

	toggle_flag(&conf,usr,INFO_BOOST);
} /* }}} */

/* cmd_jq {{{ */
void cmd_jq ( userrec_t *usr)
{
	//can be used only from console
	if (!usr->cons) return;

	pubmsg(usr,"========================================");
	pubmsg(usr,"	JQ");
	pubmsg(usr,"========================================");

	toggle_flag(&conf,usr,SHOW_JQ);
} /* }}} */

/* cmd_quit {{{ */
void cmd_quit ( userrec_t *usr)
{

	//can be used only from console
	if (!usr->cons) return;
	if ( usr->perm && !strchr(usr->perm,'c')) return;

	pubmsg(usr,"========================================");
	pubmsg(usr,"	QUIT");
	pubmsg(usr,"========================================");
	user_set_state(usr,STATE_QUIT);
} /* }}} */

/* cmd_who {{{ */
void cmd_who ( userrec_t *usr)
{
	int i;
	char str[32];
	
	if (!usr->cons) return;
	if ( usr->perm && !strchr(usr->perm,'c')) return;
	pubmsg(usr,"========================================");
	pubmsg(usr,"	WHO");
	pubmsg(usr,"========================================");

	pubmsg(usr,"USER				 FROM			 IDLE");
	for (i=0;i<NUSERS;i++)
		if (user[i] && user_tst_state(user[i],STATE_REGISTERED) &&(user[i]->cons))
		{
			uptime(user[i]->idle,str);
			sendf(usr,"%-20s %-15s %s\r\n",user[i]->nick,user[i]->ip,str);
		}
	pubmsg(usr,"========================================");
} /* }}} */

/* cmd_users {{{ */
void cmd_users ( userrec_t *usr)
{
	long i,tmp=0;

	if (!usr->cons) return;

	if ( usr->perm && !strchr(usr->perm,'c')) return;

	pubmsg(usr,"========================================");
	pubmsg(usr,"	USERS");
	pubmsg(usr,"========================================");

	for (i=0;i<NUSERS;i++)
	{
		if (user[i] && user_tst_state(user[i],STATE_LOGGEDIN|STATE_REGISTERED))
		{
			pubmsg(usr,"%-26s",user[i]->nick);
			tmp=(tmp+1)%4;

			if (!tmp) pubmsg(usr," ");
		}

	}

	if (tmp)
	{
		pubmsg(usr," ");
		tmp=0;
	}

	pubmsg(usr,"========================================");
} /* }}} */

/* cmd_kick {{{ */
void cmd_kick ( userrec_t *usr,char *line)
{
	char *tmp1;
	int i;
	char *__strtok_temp__=NULL;
	//checking if user has apriopriate permissions
	if ( usr->perm && !strchr(usr->perm,'s') && !strchr(usr->perm,'a')) return;

	if (!line||!*line) return;

	tmp1=strtok_r(line," ",&__strtok_temp__);
	pubmsg(usr,"========================================");
	pubmsg(usr,"	KICK: %s",tmp1);
	pubmsg(usr,"========================================");

	i=cmd_nick2id(usr,tmp1);

	if(i<0)return;

	if (user[i]==usr)
	{
		pubmsg(usr,"	It's simplier to use 'quit'.");
		pubmsg(usr,"========================================");
		return;
	}

    if (opbreak(usr,user[i],"kick"))return;
	
	tmp1=strtok_r(NULL,"",&__strtok_temp__);
	pubmsg(user[i],"You have been kicked off by %s.",usr->nick);
	if (tmp1) pubmsg(user[i],"Reason: %s.",tmp1);

	debug(DEBUG_STD,"Kicked off %s@%s by %s.",user[i]->nick,user[i]->ip,usr->nick);
	user_set_state(user[i],STATE_QUIT);

} /* }}} */

/*
void command(*usr,*line)

.description:
 runs the in main chat or console given command, depending on usr->cons

.args:
	userrec_t *usr - userrec of the user, who gives the command
	char *line - the line following the command

{{{*/

void command( userrec_t *usr,char *line)
{
/* const decl. {{{ */
	const char *cmd[]=
	{
		"#help",			//0
		"#password",
		"#uptime",
		"#redirect",
		"#redirect_all",
		"#version",			//5
		"#lspunished",
		"#punish",
		"#amnesty",
		"#msg",
		"#msg_all",		//10
		"#amnesty_all",
		"#conf",
		"#rehash",
		"#quit",
		"#stat",			//15
		"#whois",
		"#who",
		"#users",
		"#kick",			
		"#jq",				//20
		"#boost",
		"#mmsg",
		"#mmsg_all",
		"#pubchat",			
		"#whoip",			//25
		"#rules",
		"#msg_prv",
		"#msg_prv_all",
		"" /* indicates end of list */
	};

	const char *sep=" ";
 /* }}} */

	int cmdcount=0;
	char *cmdtmp=NULL; //stores the command
	char *cmdparam=NULL; //stores the parameters
	char *line_copy=NULL; //stores the copy of line
	char *__strtok_temp__=NULL;
/* parase the line {{{*/

	my_duplicate(line,&line_copy);

	//get the command
	cmdtmp=strtok_r(line_copy,sep,&__strtok_temp__);
	if (!usr->cons) cmdtmp=strtok_r(NULL,sep,&__strtok_temp__);

	//get the parameters of the command
	cmdparam=strtok_r(NULL,"",&__strtok_temp__);

	cmdcount=-1;
	if (cmdtmp && (*cmdtmp=='#'||usr->cons) && *cmdtmp!=0)
	{
		cmdcount=0;
		while(*cmd[cmdcount])
		{
			if (!strcasecmp(cmdtmp,cmd[cmdcount]+(usr->cons?1:0))) break;
			else cmdcount++;
		}
		if (!usr->cons && *cmd[cmdcount]) sendf(usr,"%s|",line);
	}

/* to uncomment add // at beggining
//DEBUG PURPOSES {{{
	pubmsg(usr,"{+ + + + + + + + + + + + + + + + + + + + + + + + + + + + }");
	pubmsg(usr,"	cons: %d; perm: %s;",usr->cons,usr->perm);
	pubmsg(usr,"	line: '%s'",line);
	pubmsg(usr,"	command: '%s'; parameters:'%s'",cmdtmp,cmdparam);
	pubmsg(usr,"{+ + + + + + + + + + + + + + + + + + + + + + + + + + + + }");
//DEBUG PURPOSES }}} */

/* }}} */

	switch (cmdcount)
	{
		case CMD_HELP			:cmd_help(usr,cmdparam);		goto leave;
		case CMD_PASSWORD		:cmd_password(usr,cmdparam);	goto leave;
		case CMD_UPTIME			:cmd_uptime(usr);				goto leave;
		case CMD_RULES			:cmd_rules(usr);				goto leave;
		case CMD_REDIRECT		:cmd_redirect(usr,cmdparam);	goto leave;
		case CMD_REDIRECT_ALL	:cmd_redirect_all(usr,cmdparam);goto leave;
		case CMD_VERSION		:cmd_version(usr);				goto leave;
		case CMD_LSPUNISHED		:cmd_lspunished(usr);			goto leave;
		case CMD_PUNISH			:cmd_punish(usr,cmdparam);		goto leave;
		case CMD_MSG			:cmd_msg(usr,cmdparam);			goto leave;
		case CMD_MSG_ALL		:cmd_msg_all(usr,cmdparam);		goto leave;
		case CMD_AMNESTY		:cmd_amnesty(usr,cmdparam);		goto leave;
		case CMD_AMNESTY_ALL	:cmd_amnesty_all(usr);			goto leave;
		case CMD_CONF			:cmd_conf(usr);					goto leave;
		case CMD_REHASH			:cmd_rehash(usr);				goto leave;
		case CMD_QUIT			:cmd_quit(usr);					goto leave;
		case CMD_STAT			:cmd_stat(usr);					goto leave;
		case CMD_WHOIS			:cmd_whois(usr,cmdparam);		goto leave;
		case CMD_WHO			:cmd_who(usr);					goto leave;
		case CMD_USERS			:cmd_users(usr);				goto leave;
		case CMD_KICK			:cmd_kick(usr,cmdparam);		goto leave;
		case CMD_JQ				:cmd_jq(usr);					goto leave;
		case CMD_BOOST			:cmd_boost(usr);				goto leave;
		case CMD_MMSG			:cmd_mmsg(usr,cmdparam);		goto leave;
		case CMD_MMSG_ALL		:cmd_mmsg_all(usr,cmdparam);	goto leave;
		case CMD_PUBCHAT		:cmd_pubchat(usr,cmdparam);		goto leave;
		case CMD_WHOIP			:cmd_whoip(usr,cmdparam);		goto leave;
		case CMD_MSG_PRV		:cmd_msg_prv(usr,cmdparam);		goto leave;
		case CMD_MSG_PRV_ALL	:cmd_msg_prv_all(usr,cmdparam);	goto leave;


		default:
			if(user_tst_penalty(usr,PENALTY_NOPBLCHAT) || (usr->cons && !(usr->cons&2))) goto leave;

			if (conf.max_chat_length>0 && !usr->cons && !strchr(usr->perm,'s') && strlen(line)>conf.max_chat_length) 
			{
				if (conf.kick_max_chat_length>0) 
				{
					pubmsg(usr, "You are kicked because the message sent by you was longer than %d",conf.max_chat_length);
					user_set_state(usr,STATE_QUIT);
				}
		
				pubmsg(usr,"Message was not sent because it is longer than %d",conf.max_chat_length);

				goto leave;
			}

			if(conf.allow_chat)
			{
				if (cmdtmp && *cmdtmp=='#') sendf(usr,"%s|",line);
				else
					if (usr && usr->cons&2)
					{
						sendf(NULL,"<%s> %s|",usr->nick,line);
						debug(DEBUG_CHAT,"<%s> %s",usr->nick,line);
					}
					else
						{
							sendf(NULL,"%s|",line);
							debug(DEBUG_CHAT,"%s",line);
						}
			}
			else pubmsg(usr,"Public chat is not allowed.");
	}

leave:
	my_free(line_copy);
}
/*}}}*/
