/* $Id: main.c,v 1.416 2005/11/25 11:45:03 rav Exp $ */

/* Intro {{{
 * ----------------------------------------------------------------
 * DConnect Daemon
 *
 *	 #(@) Copyright (c) 2002, DConnect development team
 *	 #(@) Homepage: http://www.dc.ds.pg.gda.pl/
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
 * MA 02139, USA.
 *
 * ----------------------------------------------------------------
 * }}} */

/* include local header files {{{ */
#include "pch.h"
#include "main.h"
#include "conf.h"
#include "misc.h"
#include "net.h"
#include "match.h"
#include "penalties.h"
#include "console.h"
#include "cmd.user.h"

/* }}} */

/* reserved_nicks[] - nicks that can't be owned by normal users {{{ */
const char *reserved_nicks[]=
{
	"Hub",
	"Hub-Security",
	"Vandel\\Debug",
	"Client",
	"" /* indicates end of list */
}; /* }}} */

typedef union{
		void *voidp;
		int integer;
}int_voidp_t;

int NPENAL; /* penalties count */
time_t last_penalties_update=0;

int NUSERS; /* user count */
int maxfd;

int rehash=0; // determines if #rehash command was used

Tpenalty *penalties[MAXPENALTIES]; /* table of penalties */
userrec_t *user[MAXUSERS];	/* users table */
config_t conf;		/* configuration data */

pthread_t listen_th_main[MAX_N_LISTEN_MAIN];	/* listen thread handle for main service */
pthread_t th_user_manager[MAX_N_USER_MANAGER] ;/* handles managing user's communicates */

pthread_t listen_th_cons;	/* listen thread handle for remote console */
pthread_t listen_th_sr;		/* listen thread handle for minslots checking */
pthread_t th_cleaner;

struct sockaddr_in local_main,local_sr, local_cons;

so_opts_t listen_opts,user_opts;

int signal_p=0;

pthread_mutex_t mutex;		/* provides synchronization between threads */
pthread_mutex_t mutex_inet_ntoa; // securing inet_ntoa
pthread_mutex_t mutex_my_hosts_access; // securing my_hosts_access


pthread_mutex_t mutex_listen_main[MAX_N_LISTEN_MAIN];
pthread_mutex_t mutex_user_manager[MAX_N_USER_MANAGER];

pthread_mutex_t mutex_is_managed; // secures that user is served only by one thread
int managed[MAX_N_USER_MANAGER]; // sets which thread is working on which user

pthread_mutex_t mutex_th_cleaner;

pthread_mutex_t mutex_penalties;

pthread_mutex_t mutex_listen_cons;
pthread_mutex_t mutex_listen_sr;

pthread_mutex_t mutex_my_send;

pthread_mutex_t mutex_myinfo;



int listen_main;			/* main listening socket */
int listen_cons;			/* console listening socket */
int listen_sr;				/* minslots checking socket */
char DAEMON;				/* set if running in background */
time_t starttime;			/* start time in sec */

/* logging */
FILE* file_net;
FILE* file_proto;
FILE* file_wrng;
FILE* file_con;
FILE* file_chat;

/* these are for tcpd */
int allow_severity;
int deny_severity;

int test_penalty_counter=0;

#ifdef HAVE___PROGNAME
extern char *__progname;
#else
char *__progname;
#endif

void lock_threads()
{
	int i;

	pthread_mutex_lock(&mutex_th_cleaner);
	pthread_mutex_lock(&mutex_listen_sr);
	for(i=0;i<MAX_N_LISTEN_MAIN;i++) pthread_mutex_lock(&mutex_listen_main[i]);
	pthread_mutex_lock(&mutex_listen_cons);
	for(i=0;i<MAX_N_USER_MANAGER;i++) pthread_mutex_lock(&mutex_user_manager[i]);
}

void unlock_threads()
{
	int i;

	for(i=0;i<MAX_N_USER_MANAGER;i++) pthread_mutex_unlock(&mutex_user_manager[i]);
	pthread_mutex_unlock(&mutex_listen_cons);
	for(i=0;i<MAX_N_LISTEN_MAIN;i++) pthread_mutex_unlock(&mutex_listen_main[i]);
	pthread_mutex_unlock(&mutex_listen_sr);
	pthread_mutex_unlock(&mutex_th_cleaner);
}

void make_sockets()
{

	/* setup and bind the listening socket */
	if ((listen_main=socket(AF_INET,SOCK_STREAM,0))==-1)
	{
		debug(DEBUG_SYS|DEBUG_STD,0,"Socket error - terminating: %s.", strerror(errno));
		exit(0);
	}

	if (setsockopts(listen_main,&listen_opts)) debug(DEBUG_WRNG,0,"setsockopts() failed.");

	memset(&local_main,0,sizeof(struct sockaddr_in));
	local_main.sin_family=AF_INET;
	local_main.sin_port=htons(conf.listen_port_main);
	inet_aton(conf.listen_interface,&local_main.sin_addr);

	if (bind(listen_main,(struct sockaddr *)&local_main,sizeof(struct sockaddr))<0)
	{
		debug(DEBUG_SYS|DEBUG_STD,"Can't bind the main listening socket - terminating: %s.", strerror(errno));
		exit(0);
	}

	if (listen(listen_main,SOMAXCONN)<0)
	{
		debug(DEBUG_SYS|DEBUG_STD,"listeN() failed - terminating: %s.", strerror(errno));
		exit(0);
	}

	if (conf.minslots)
	{
		/* setup and bind the listening socket */
		if ((listen_sr=socket(AF_INET,SOCK_DGRAM,0))==-1)
		{
			debug(DEBUG_SYS|DEBUG_STD,0,"Socket error - terminating: %s.", strerror(errno));
			exit(0);
		}

		if (setsockopts(listen_sr,&listen_opts)) debug(DEBUG_WRNG,0,"setsockopts() failed.");

		memset(&local_sr,0,sizeof(struct sockaddr_in));
		local_sr.sin_family=AF_INET;
		local_sr.sin_port=htons(conf.listen_port_sr);

		inet_aton(conf.listen_interface,&local_sr.sin_addr);

		if (bind(listen_sr,(struct sockaddr *)&local_sr,sizeof(struct sockaddr))<0)
		{
			debug(DEBUG_SYS|DEBUG_STD,"Can't bind the minslots checking listening socket - terminating: %s.", strerror(errno));
			exit(0);
		}

	}

	if (conf.listen_port_cons)
	{
		/* setup and bind the remote console listening socket */
		if ((listen_cons=socket(AF_INET,SOCK_STREAM,0))==-1)
		{
			debug(DEBUG_SYS|DEBUG_STD,"Socket error - console disabled: %s.", strerror(errno));
			exit(1);
		}

		if (setsockopts(listen_cons,&listen_opts)) debug(DEBUG_WRNG,"setsockopts() failed.");

		memset(&local_cons,0,sizeof(struct sockaddr_in));

		local_cons.sin_family=AF_INET;
		local_cons.sin_port=htons(conf.listen_port_cons);
		inet_aton(conf.listen_interface,&local_cons.sin_addr);

		if (bind(listen_cons,(struct sockaddr *)&local_cons,sizeof(struct sockaddr))<0)
		{
			debug(DEBUG_SYS|DEBUG_STD,"Can't bind the console listening socket - console disabled: %s.", strerror(errno));
			exit(1);
		}

		if (listen(listen_cons,5)<0)
		{
			debug(DEBUG_SYS|DEBUG_STD,"listeN() for console failed - terminating: %s.", strerror(errno));
			exit(0);
		}
	}
}

int user_set_state(userrec_t *usr, int state)
{
	//if user quits his state cannot be changed
	if (user_tst_state(usr,STATE_QUIT)) return 0;

	if (usr->nick) debug(DEBUG_PROTO,"'%s' %d -> %d.",usr->nick,usr->state,state);
	
	if (state==STATE_QUIT)
	{    
	    if (user_tst_state(usr,STATE_REGISTERED|STATE_LOGGEDIN)) usr->state=STATE_QUIT_SHOW;
	    else usr->state=STATE_QUIT;
	}
	else usr->state=state;

	return 0;
}

int user_tst_state(userrec_t *usr, int state)
{

	return ((usr->state)&state)>0;
}

/* adduser() - allocate and initialize new DC user {{{ */
void adduser(int sock,char *remote,char *local)
{
	char lock[101], pk[17];
	userrec_t *tmp=NULL,tmpik;

//	debug(DEBUG_PROTO,"adduser() enter");

	if (setsockopts(sock,&user_opts)) debug(DEBUG_WRNG,"setsockopts() failed.");

	/* we use select() anyway, but let's care for blocking send() */
	if (fcntl(sock,F_SETFL,O_NONBLOCK)<0) debug(DEBUG_WRNG,"fcntl() failed: %s.", strerror(errno));

	/* testing if user is listed in hosts.allow */
	if (denied(sock))
	{
		tmpik.sock=sock;
		tmpik.cons=0;
		tmpik.nick=NULL;
		tmpik.ip=remote;

		sendf(&tmpik,"$Lock AAAAAAAAAABBBBBBBBBB Pk=01234|$HubName %s|",conf.hubname);

		pubmsg(&tmpik,"Contact with %s",conf.admin_contact);
		pubmsg(&tmpik,"%s","You are not allowed to connect.");

		if (conf.redirect_switch&REDIRECT_SWITCH_ACCESS)
		{
			pubmsg(&tmpik,"You are being redirected to %s",conf.redirect_access);
			sendf(&tmpik,"$ForceMove %s|",conf.redirect_access);

			debug(DEBUG_NET,"Host %s is not allowed to connect and is being redirected.",remote);
		}
		else debug(DEBUG_NET,"Host %s is not allowed to connect.",remote);

		my_free(remote);
		my_free(local);
		close(sock);
		goto leave;
	}

	tmp=(userrec_t *)my_malloc(sizeof(userrec_t));
	memset(tmp,0,sizeof(userrec_t));

	user_set_state(tmp,STATE_WELCOME);

	/* getting user's ip addresses */
	tmp->ip=remote;
	tmp->con_ip=local;
	tmp->nick=NULL;

	tmp->buf=NULL;

	tmp->sock=sock;
	tmp->slots=0;
	tmp->slots_old=0;

	/* client complains that it couldn't find $Lock when msgs are sent before it */
	lockgen(lock,pk);

	sendf(tmp,"$Lock %s Pk=%s|$HubName %s|",lock,pk,conf.hubname);

	if((NUSERS>=conf.userlimit)||(NUSERS+1>=MAXUSERS)) /* '+1' reserves 1 slot for admin console */
	{
		pubmsg(tmp,"Sorry, this Hub is currently full!");

		if (conf.redirect_switch&REDIRECT_SWITCH_HUB_IS_FULL)
		{
			pubmsg(tmp,"You are being redirected to %s",conf.redirect_hub_is_full);
			sendf(tmp,"$ForceMove %s|",conf.redirect_hub_is_full);
		}

		disconnect(tmp,NULL);
		goto leave;
	}

	tmp->idle=time(NULL);
	tmp->last_search=0;

	tmp->perm=NULL;
	tmp->password=NULL;

	pthread_mutex_lock( &mutex );
	user[NUSERS++]=tmp;
	pthread_mutex_unlock( &mutex );

leave:
//	debug(DEBUG_PROTO,"adduser() leave");
	return;
} /* }}} */

/* deluser() - free normal or console user {{{ */
void deluser(userrec_t *usr,char *reason)
{
	int id=usr2id(usr);

	if (!usr) return;

	user[id]=user[--NUSERS];
	user[NUSERS]=NULL;

	if (user_tst_state(usr,STATE_SHOW_QUIT))
	{
		sendf(NULL,"$Quit %s|",usr->nick);
		if (conf.flags & SHOW_JQ) sendf(NULL,"## %s has quit|",usr->nick);
	}
	disconnect(usr,reason);

} /* }}} */

/* listen_thread_maiN() - listen for new connections to main service {{{ */
void listen_thread_main(void *voidp)
{
	unsigned int new_sock,remote_size,size_addr=sizeof(struct sockaddr_in);
	struct sockaddr_in remote,addr;
	int No;
	int_voidp_t parameter;

	char *remote_ip,*local_ip;

	parameter.voidp=voidp;
	No=parameter.integer;

	debug(DEBUG_PROTO,"listen_thread_main(%d)",No);

	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS,NULL);

	remote_size=sizeof(struct sockaddr_in);
	/* accept connections */
	while(1)
	{

		if ((new_sock=accept(listen_main,(struct sockaddr *)&remote,&remote_size))<0)
		{
			if (errno!=EINTR) debug(DEBUG_WRNG,"accept() failed: %s.", strerror(errno));
			continue;
		}

		pthread_mutex_lock(&mutex_listen_main[No]);

		getsockname(new_sock,(struct sockaddr *)&addr,&size_addr);
		debug(DEBUG_NET,"Connection to DC from %s:%d to %s.",remote_ip=my_inet_ntoa(remote.sin_addr),
															 ntohs(remote.sin_port),
															 local_ip=my_inet_ntoa(addr.sin_addr));
		adduser(new_sock,remote_ip, local_ip);
		pthread_mutex_unlock(&mutex_listen_main[No]);

	}

	debug(DEBUG_WRNG,"listen_thread_main: this should be never reached!");
	pthread_exit(0); /* never reached */

} /* }}} */

int penalty_welcome(userrec_t *usr)
{
	int time_, days, hours,sec,min;

	/*sOP cannot be punished*/
	if (usr->perm && strchr(usr->perm,'s')) return 0;

	/* get the penalties for the user*/
	user_penalty(usr);

	if (user_tst_penalty(usr,~PENALTY_NONE))
	{
		time_=user_exp_penalty(usr->penalty);

		days=time_/(long)(24*60*60);
		time_=time_%(24*60*60);

		hours=time_/(long)(60*60);
		time_=time_%(60*60);

		min=time_/(long)60;
		time_=time_%60;

		sec=time_;
		pubmsg(usr,"Your host(%s@%s) is punished with:",usr->penalty->nick,usr->penalty->ip);
		
		pubmsg(usr,"%s%s%s%s%s",(usr->penalty->penalty&PENALTY_BANNED)?"[ban] ":"",
					(usr->penalty->penalty&PENALTY_NODL)?"[nodl] ":"",
					(usr->penalty->penalty&PENALTY_NOSEARCH)?"[nosearch] ":"",
					(usr->penalty->penalty&PENALTY_NOPBLCHAT)?"[nopblchat] ":"",
					(usr->penalty->penalty&PENALTY_NOPRVCHAT)?"[noprvchat] ":"");

		pubmsg(usr,"Time left: %ld days and %ld:%ld:%ld",days,hours,min,sec);
		pubmsg(usr,"REASON: %s",usr->penalty->reason);
		pubmsg(usr,"-----------------------------------");

	}


	if (user_tst_penalty(usr,PENALTY_BANNED)) 
	{
		debug(DEBUG_NET,"'%s' trying to connect from baned %s",usr->nick,usr->ip);
	    user_set_state(usr,STATE_QUIT);
	    return 1;
	}

	return 0;
}
// this function is used to send nicklist to a user
void GetNickList(userrec_t *usr)
{
	char ops[((NICK_LEN+2)*MAXUSERS+2)],
	     nicks[((NICK_LEN+2)*MAXUSERS+2)];
	int i;
		
		ops[0]='\000';
		nicks[0]='\000';

    	strcat(ops,"HUB$$");
		strcat(nicks,"HUB$$");
	
//	debug(DEBUG_PROTO,"GetNicklist( %s ) enter",usr->nick);
	for(i=0;i<NUSERS;i++)
		if (user[i] && user_tst_state(user[i],STATE_REGISTERED))
		{
//			debug(DEBUG_PROTO,"%d %d ",i,NUSERS);
			strcat(nicks,user[i]->nick);
			strcat(nicks,"$$");

			if (user[i]->perm && (strchr(user[i]->perm,'s') || strchr(user[i]->perm,'a')))
			{
				strcat(ops,user[i]->nick);
				strcat(ops,"$$");
			}
		}

//	debug(DEBUG_PROTO,"nicks %s",nicks);
//	debug(DEBUG_PROTO,"ops %s",ops);

	//if user is not registered
	if (!user_tst_state(usr,STATE_REGISTERED))
	{
		strcat(nicks,usr->nick);
		strcat(nicks,"$$");

		if (usr->perm && (strchr(usr->perm,'a')||strchr(usr->perm,'s')))
		{
			strcat(ops,usr->nick);
			strcat(ops,"$$");
		}
	}

//	debug(DEBUG_PROTO,"nicks %s",nicks);
//	debug(DEBUG_PROTO,"ops %s",ops);

	sendf(usr,"$NickList %s|",nicks);

//	dc_myinfo(NULL,usr);
//	dc_myinfo(usr,usr);							
	
	if (strlen(ops))
	{
		if (usr->perm && (strchr(usr->perm,'s') || strchr(usr->perm,'a')))
		{
			sendf(usr,"$OpList %s|",ops);
			sendf(NULL,"$OpList %s|",ops);
		}
		else sendf(usr,"$OpList %s|",ops);
	}

//	debug(DEBUG_PROTO,"GetNicklist( %s ) leave",usr->nick);
}

// this function is used to chect the minslots
void listen_thread_SR(void *args)
{
	unsigned int remote_size;
	int i=0,n;
	struct sockaddr_in remote;
	userrec_t *usr=NULL;
	char *ip=NULL, bufor[10001], *cmd=NULL, *nick=NULL, *s_slots=NULL, *__strtok_temp__=NULL, nick_prev[NICK_LEN], *filename;

	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS,NULL);

	nick_prev[0]=0;
	
	remote_size=sizeof(struct sockaddr_in);
	/* accept connections */
	while(1)
	{
		n=recvfrom(listen_sr,bufor,10000,0,(struct sockaddr*)&remote,&remote_size);		
		pthread_mutex_lock(&mutex_listen_sr);

		if (n<=0) goto next2;

		bufor[n]=0;
//		debug(DEBUG_PROTO,"SLOTS %s",bufor);

		cmd=strtok_r(bufor," ",&__strtok_temp__);
		if (!cmd || strcmp(cmd,"$SR")) goto next2;

		nick=strtok_r(NULL," ",&__strtok_temp__);
		if (!nick) goto next2;

		filename=strtok_r(NULL,"\005",&__strtok_temp__);
		
		i++;
		i%=11;
		if (!filename || (i && strcmp(filename,".") && !strcasecmp(nick,nick_prev))) goto next2;
		
		if (!i)nick_prev[0]=0;		
		else strcpy(nick_prev,nick);
		
		s_slots=strtok_r(NULL,"/",&__strtok_temp__);
		if (!s_slots) goto next2;

		s_slots=strtok_r(NULL,"\005",&__strtok_temp__);
		if (!s_slots) goto next2;
//		debug(DEBUG_PROTO,"SLOTS TESTING");


		ip=my_inet_ntoa(remote.sin_addr);

		n=nick2id(nick);
		if (n<0) goto next;

		usr=user[n];

		if (!user_tst_state(usr,STATE_SR) || 
			strchr(usr->perm,'s') || 
			strchr(usr->perm,'r') || 
			(usr->ip && !match(usr->ip,ip))) goto next;

		usr->slots=strtol(s_slots, &__strtok_temp__, 10);
		user_set_state(usr,(usr->state^STATE_SR)|STATE_SLOTS_TESTED);
	
	next:
		my_free(ip);

	next2:
		pthread_mutex_unlock(&mutex_listen_sr);
	}

	debug(DEBUG_WRNG,"listen_thread_main: this should be never reached!");
	pthread_exit(0); /* never reached */
} /* }}} */

/* finish() - signal handler and program termination {{{ */
void finish(int sig)
{
	int i;

	if ((sig!=SIGKILL || sig!=SIGSEGV) && signal_p) return;
	signal_p=1;

	if (sig!=SIGPIPE) debug(DEBUG_STD|DEBUG_PROTO,"SIG%s%s%s%s%s%s received - %s%s%s%s",	
																		sig==SIGHUP?"HUP":"",
																		sig==SIGKILL?"KILL":"",
																		sig==SIGTERM?"TERM":"",
																		sig==SIGQUIT?"QUIT":"",
																		sig==SIGSEGV?"SEGV":"",
																		sig==SIGINT?"INT":"",
																		
																		sig==SIGHUP?"reconfiguring":"",
																		sig==SIGSEGV?"crashing":"",
																		sig==SIGKILL?"fast terminating":"",
																		(sig==SIGTERM||sig==SIGQUIT||sig==SIGINT)?"cleaning and terminating":"");
																		
	switch(sig)
	{
		/* rehash on SIGHUP */
		case SIGHUP:
					/* blocking working threads */

					lock_threads();

					parse_config(0,&conf);
					make_paths(&conf);

					if ((*(conf.log_main))&&(conf.log_level&DEBUG_STD) && freopen(conf.log_main,"a",stderr)==NULL) debug(DEBUG_SYS,"Couldn't open main log file (%s): %s.",conf.log_main,strerror(errno));

					if (file_net)
					{
						fclose(file_net);
						file_net=NULL;
					}

					if ((*(conf.log_net))&&(conf.log_level&DEBUG_NET) && (file_net=fopen(conf.log_net,"a"))==NULL) debug(DEBUG_SYS|DEBUG_STD,"Couldn't open network access log file (%s): %s.",conf.log_net,strerror(errno));

					if (file_proto)
					{
						fclose(file_proto);
						file_proto=NULL;
					}

					if ((*(conf.log_proto))&&(conf.log_level&DEBUG_PROTO) && (file_proto=fopen(conf.log_proto,"a"))==NULL) debug(DEBUG_SYS|DEBUG_STD,"Couldn't open protocol log file (%s): %s.",conf.log_proto,strerror(errno));

					if (file_wrng)
					{
						fclose(file_wrng);
						file_wrng=NULL;
					}

					if ((*(conf.log_wrng))&&(conf.log_level&DEBUG_WRNG) && (file_wrng=fopen(conf.log_wrng,"a"))==NULL) debug(DEBUG_SYS|DEBUG_STD,"Couldn't open warnings log file (%s): %s.",conf.log_wrng,strerror(errno));

					if (file_con)
					{
						fclose(file_con);
						file_con=NULL;
					}

					if ((*(conf.log_con))&&(conf.log_level&DEBUG_CON) && (file_con=fopen(conf.log_con,"a"))==NULL) debug(DEBUG_SYS|DEBUG_STD,"Couldn't open console log file (%s): %s.",conf.log_con,strerror(errno));

					if (file_chat)
					{
						fclose(file_chat);
						file_chat=NULL;
					}

					if ((*(conf.log_chat))&&(conf.log_level&DEBUG_CHAT) && (file_chat=fopen(conf.log_chat,"a"))==NULL) debug(DEBUG_SYS|DEBUG_STD,"Couldn't open chat log file (%s): %s.",conf.log_chat,strerror(errno));

					unlock_threads();

					signal_p=0;
					debug(DEBUG_PROTO|DEBUG_STD,"Daemon Reconfigured");
					return;

		case SIGSEGV:
					openlog("dcd",LOG_PID|LOG_CONS,LOG_DAEMON);
					syslog(LOG_CRIT,"Segmentation fault. Crashing!\n");

					close(listen_main);
					if (conf.listen_port_cons) close(listen_cons);
					if (conf.minslots)close(listen_sr);					

					closelog();
					debug(DEBUG_PROTO|DEBUG_STD,"Daemon Crashed");
					exit(SIGSEGV);

		case SIGKILL:
					debug(DEBUG_PROTO|DEBUG_STD,"Daemon terminated fast");
					close(listen_main);
					if (conf.listen_port_cons) close(listen_cons);
					if (conf.minslots)close(listen_sr);					
					exit(SIGKILL);

		case SIGTERM:
		case SIGINT:
		case SIGQUIT:
					lock_threads();

				/* normaly this makes dead lock of thread when then trying to start new thread,
				 * but in this case we shutdown the hub so it is ok */

					debug(DEBUG_PROTO|DEBUG_STD,"Threads cancelling");
					for(i=0;i<conf.n_listen_main;i++) pthread_cancel(listen_th_main[i]);
					for(i=0;i<conf.n_user_manager;i++) pthread_cancel(th_user_manager[i]);

					if (conf.minslots) pthread_cancel(listen_th_sr);
					if (conf.listen_port_cons) pthread_cancel(listen_th_cons);
					pthread_cancel(th_cleaner);
					debug(DEBUG_PROTO|DEBUG_STD,"Threads cancelled");

					for (i=0;i<NUSERS;i++) disconnect(user[i],"Hub shutdown!");

					close(listen_main);
					if (conf.listen_port_cons) close(listen_cons);
					if (conf.minslots)close(listen_sr);					

					debug(DEBUG_PROTO|DEBUG_STD,"Daemon Terminated");
					exit(sig);

		/* ignore any other signal */
		default:
				signal_p=0;
				return;
	}
} /* }}} */

/* buf2line() - extract single request line from buf {{{ */
char *buf2cmd(userrec_t *usr)
{
	char delim_con[3]="\n",
		 delim_cmd[2]="|",
		 *delim,
		 *line=NULL,
		 *cmd=NULL, 
		 *tmp=NULL,
		 *__strtok_temp__=NULL;
	
//	debug(DEBUG_PROTO,"buff2cmd('%s,%s') enter",usr->nick,usr->buf);
//	debug(DEBUG_PROTO,"buff2cmd() enter");

	if (!usr->buf || !*(usr->buf)) goto leave;

	if (usr->cons) delim=delim_con;
	else delim=delim_cmd;
	
	line=strtok_r(usr->buf,delim,&__strtok_temp__);
	
	if (line && *line!='\r')
	{
		my_duplicate(line,&cmd);
		if (__strtok_temp__ && *__strtok_temp__) my_duplicate(__strtok_temp__,&tmp);
		my_free(usr->buf);
		usr->buf=tmp;
	} 

	if (usr->cons && user_tst_state(usr, STATE_KEY|STATE_PASSWORD) && !cmd) user_set_state(usr, STATE_QUIT);

leave:
//	debug(DEBUG_PROTO,"buff2cmd() leave");
//	debug(DEBUG_PROTO,"buff2cmd('%s','%s') leave",usr->nick,cmd);
	return cmd;
} /* }}} */

void msg2buff(userrec_t *usr, char *message)
{
	char *tmp;
	int len=1;

//	debug(DEBUG_PROTO,"msg2buff('%s', '%s') enter",usr->buf,message);
//	debug(DEBUG_PROTO,"msg2buff() enter");
	if (!message || !*message) goto leave;
	
	if (usr->buf) len+=strlen(usr->buf);

	len+=strlen(message);
	tmp=my_malloc(len);
	tmp[0]=0;
	
	if (usr->buf) strcat(tmp,usr->buf);
	strcat(tmp,message);
	
	my_free(usr->buf);
	usr->buf=tmp;
leave:
//	debug(DEBUG_PROTO,"msg2buff('%s') leave",usr->buf);
//	debug(DEBUG_PROTO,"msg2buff() leave");
	return;
}



/* manage_sock() - proces requests remaining on a socket {{{ */
void manage_sock( userrec_t *usr)
{
	const char *commands[]=
	{
		"Key",
		"Quit",
		"ValidateNick",
		"GetNickList",
		"MyINFO",
		"GetINFO",
		"Kick",
		"MyPass",
		"Search",
		"MultiSearch",
		"To:",
		"Version",
		"ConnectToMe",
		"MultiConnectToMe",
		"RevConnectToMe",
		"UniSearch",
		"SR",
		"OpForceMove",
		"Supports",
		"" /* indicates end of list */
	};

	const char *sep=" ";
	char *line=NULL, *line_copy=NULL, *cmd=NULL, *par=NULL;
	char *tmp1=NULL, *tmp2=NULL,*tmp3=NULL,*__strtok_temp__=NULL;
	int cNO=0,i=0,counter=1;
	myinfo_t myinfo_tmp;


//	debug(DEBUG_PROTO,"manage_sock(%s) enter",usr->nick);

	while(counter && (line=(char *)buf2cmd(usr))) //makes a line from buf
	{
		my_duplicate(line,&line_copy);

		counter++;
		counter=counter%conf.max_commands;
		
		if (usr->cons)
		{
			if (manage_cons(usr,line_copy)) goto leave; //
			else goto end;
		}

		debug(DEBUG_PROTO,"RECV %s '%s'",usr->nick==NULL?"Unknown":usr->nick,line);

		if (!(cmd=strtok_r(line_copy,sep,&__strtok_temp__))) goto end;
		if (!*cmd) goto end;
		par=strtok_r(NULL,sep,&__strtok_temp__);
		cNO=0; /* - $ command number*/

		/* Parse commands {{{ */
		if (*cmd=='$')
		{
			while(*commands[cNO])
			{
				if (!strcasecmp(&cmd[1],commands[cNO])) break;
				else cNO++;
			}
		}
		else
			if ((*cmd=='<')&&(cmd[strlen(cmd)-1]=='>')) cNO=-1;
			else cNO=-2; // unrecognized command

		if ((user_tst_state(usr,STATE_CONNECTED) && cNO!=0 && cNO!=18) ||
			(user_tst_state(usr,STATE_KEY) && cNO!=2 ))
		{
			user_set_state(usr,STATE_QUIT);
			pubmsg(usr,"Broken registration.");
			goto leave;
		}

		switch(cNO)
		{
			/* -1. pubmsg {{{ */
			case -1:
					if (!user_tst_state(usr,STATE_REGISTERED)) goto end;

					cmd[strlen(cmd)-1]=0;
					cmd++;

					if (!strcasecmp(cmd,usr->nick)) command(usr,line);
					else debug(DEBUG_PROTO,"On PublicMsg from %s@%s: Fake nick not allowed. (%s)",usr->nick,usr->ip, line);

					goto end; /* }}} */

			/* 0. $Key {{{ */
			case 0:
					if (!user_tst_state(usr,STATE_CONNECTED)) goto end;
					if (!validkey(par))
					{
						pubmsg(usr,"Invalid $Key reply.");
						user_set_state(usr,STATE_QUIT);
						debug(DEBUG_PROTO,"On Key from %s@%s: Invalid reply. (%s)",usr->nick,usr->ip,line);
						goto leave;
					}

					user_set_state(usr,STATE_KEY);
					goto end; /* }}} */

			/* 1. $Quit {{{ */
			case 1:
					if (!par) goto end;
					if (strcasecmp(par,usr->nick))
					{
						debug(DEBUG_PROTO,"On quit from %s@%s: Fake nick not allowed. (%s)",usr->nick,usr->ip,line);
						goto end;
					}

					user_set_state(usr,STATE_QUIT);

					goto leave; /* }}} */

			/* 2. $ValidateNick {{{ */
			case 2:
					if (!par || user_tst_state(usr,STATE_LOGGEDIN|STATE_REGISTERED|STATE_PASSWORD))goto end;

					debug(DEBUG_NET,"'%s' from %s is validating",par,usr->ip);

					switch (validnick(usr,par))
					{
						case 0:
								user_set_state(usr,STATE_QUIT);
								debug(DEBUG_NET,"'%s' not loged in",par);
								goto leave;

						case 1:
								/* get the penalties for the user*/
								my_duplicate(par,&usr->nick);

								if (penalty_welcome(usr)) goto end;

								user_set_state(usr,STATE_LOGGEDIN);

								debug(DEBUG_NET,"'%s' logged in",usr->nick);
								sendf(usr,"$Hello %s|",usr->nick);
								goto end;

						case 2:
								my_duplicate(par,&usr->nick);

								if (penalty_welcome(usr)) goto end;

								user_set_state(usr,STATE_PASSWORD);
								debug(DEBUG_NET,"'%s' requires password",usr->nick);
								sendf(usr,"$GetPass|");
								goto end;
					}
			/* }}} */

			/* 3. $GetNickList {{{ */
			case 3:
					if (user_tst_state(usr,STATE_LOGGEDIN) && (conf.flags&INFO_BOOST)) goto end;

					GetNickList(usr);

					goto end; /* }}} */

			/* 4. $MyINFO {{{ */
			case 4:
					if(!par || !user_tst_state(usr,STATE_LOGGEDIN|STATE_REGISTERED)) goto end;
					
					myinfo_tmp.email=NULL;
					myinfo_tmp.ctype=NULL;
					myinfo_tmp.desc=NULL;

					if (fill_myinfo(usr, &myinfo_tmp, line))
					{
						if (usr->perm && !strchr(usr->perm,'s') && !strchr(usr->perm,'m') && myinfo_tmp.share<(conf.minshare*1024*1024))
						{
							pubmsg(usr,"<Hub> You must share at least %"PRIu64"MB on this hub.|",conf.minshare);

							if (conf.redirect_switch&REDIRECT_SWITCH_MINSHARE)
							{
								pubmsg(usr,"You are being redirected to %s",conf.redirect_minshare);
								sendf(usr,"$ForceMove %s|",conf.redirect_minshare);
							}

							user_set_state(usr,STATE_QUIT);
							goto leave;
						}

						if (!compare_myinfo(&myinfo_tmp,&usr->myinfo)) goto leave;
						
						move_myinfo(&myinfo_tmp, &usr->myinfo);

						if (user_tst_state(usr,STATE_LOGGEDIN))
						{
							if (conf.conf_motd && (!conf.minslots || (usr->perm && (strchr(usr->perm,'s') || strchr(usr->perm,'r'))))) 
							{
								playfile(usr,conf.conf_motd);
								pubmsg(usr,"-----------------------------------");
							}

							if (conf.flags & SHOW_JQ) sendf(NULL,"## %s has joined|",usr->nick);

							if (conf.flags & INFO_BOOST)
							for(i=0;i<NUSERS;i++)
								if (user[i] && user_tst_state(user[i],STATE_REGISTERED))
									dc_myinfo(usr,user[i]);									
						}

						if (conf.minslots && usr->perm && !strchr(usr->perm,'r') && !strchr(usr->perm,'s'))
						{
							if (!usr->slots) pubmsg(usr,"Checking for slots. Please wait.");

							user_set_state(usr,STATE_SR|usr->state);

							//wrapper for dctc and older valknut client
							sendf(usr,"$Search %s:%d F?F?0?1?fakesearch|",usr->con_ip,conf.listen_port_sr);

							//dc++, dc, valknut
							sendf(usr,"$Search %s:%d F?F?0?1?.|",usr->con_ip,conf.listen_port_sr);

							usr->idle=time(NULL);
							goto leave;
						}

						user_set_state(usr,STATE_REGISTERED);
						dc_myinfo(NULL,usr);

						goto leave;
					}

					user_set_state(usr,STATE_QUIT);
					pubmsg(usr,"Invalid MyINFO received.");
					goto leave; /* }}} */

			/* 5. $GetINFO {{{ */
			case 5:
					if (conf.flags&INFO_BOOST) goto end;

					if(!par || !user_tst_state(usr,STATE_LOGGEDIN|STATE_SR|STATE_REGISTERED)) goto end;

					i=nick2id(par);
					if (i<0)
					{
//						debug(DEBUG_PROTO,"On GetINFO from %s@%s: No such user. (%s)",usr->nick,usr->ip,line);
						goto end;
					}

					if (user[i] && user_tst_state(user[i],STATE_REGISTERED)) dc_myinfo(usr,user[i]);
					goto end; /* }}} */

			/* 6. $Kick {{{ */
			case 6:
					if(!par || !user_tst_state(usr,STATE_REGISTERED)) goto end;

					if(usr->perm && !strchr(usr->perm,'a') && !strchr(usr->perm,'s'))
					{
						debug(DEBUG_NET,"%s(not OP) tried to kick (%s)",usr->nick,par);
						pubmsg(usr,"You are Not an Op");
						goto end;
					}

					if (!strcmp(usr->nick,par))
					{
						privmsg(usr,NULL,"You are trying to kick yourself");
						goto end;
					}

					if ((i=nick2id(par))<0)
					{
						pubmsg(usr,"You are trying to kick a person, that doesn't exist");
						goto end;
					}

					if (usr->perm!=NULL && user[i]->perm!=NULL)
					{
						if (strchr(user[i]->perm,'a') && strchr(usr->perm,'a'))
						{
							privmsg(usr,NULL,"You are trying to kick OP");
							goto end;
						}

						if (strchr(user[i]->perm,'s') && (strchr(usr->perm,'a') || strchr(usr->perm,'s')))
						{
							privmsg(usr,NULL,"You are trying to kick sOP");
							goto end;
						}

						if (strchr(user[i]->perm,'a') && strchr(usr->perm,'s')) privmsg(usr,NULL,"You are kicking OP");
					}
					else goto end;

					debug(DEBUG_NET,"'%s'(OP) kicked '%s'",usr->nick,par);
					privmsg(user[i],NULL,"You were Kicked");

					user_set_state(user[i],STATE_QUIT);
					goto end; /* }}} */

			/* 7. $MyPass {{{ */
			case 7:
					if(!par || !user_tst_state(usr,STATE_PASSWORD)) goto end;

					if (!strcmp(usr->password,par))
					{
						debug(DEBUG_NET,"'%s' sent correct password", usr->nick);
						pubmsg(usr,"Your password is correct");

						if (usr->perm!=NULL)
						{
							if (strchr(usr->perm,'a'))
							{
								pubmsg(usr,"You become an OP");
								debug(DEBUG_NET,"%s becomes an OP", usr->nick);
							}

							if (strchr(usr->perm,'s'))
							{
								pubmsg(usr,"You become an sOP");
								debug(DEBUG_NET,"'%s' becomes an sOP", usr->nick);
							}

							sendf(usr,"$LogedIn|");

						}
						sendf(usr,"$Hello %s|",usr->nick);

						debug(DEBUG_NET,"'%s' is in", usr->nick);

						user_set_state(usr,STATE_LOGGEDIN);
						pubmsg(usr,"-----------------------------------");
						//wrapper for StrongDC++
//						GetNickList(usr);
					}else
						{
							debug(DEBUG_NET,"'%s' sent incorrect password", usr->nick);
							pubmsg(usr,"Your password is incorrect");
							sendf(usr,"$BadPass|$GetPass|");
							goto end;
						}

					goto end; /* }}} */

			/* 8. $Search {{{ */
			case 8:								
					if(!conf.allow_search || !par || !user_tst_state(usr,STATE_REGISTERED) || user_tst_penalty(usr,PENALTY_NOSEARCH)) goto end;

					if (!(cmd=strtok_r(par,":",&__strtok_temp__))) goto end;

					if (conf.search_interval && 
						usr->last_search && !strchr(usr->perm,'i') && !strchr(usr->perm,'s') && 
						difftime(time(NULL),usr->last_search)<(double)conf.search_interval)
						{
							if (difftime(time(NULL),usr->sent_search_interval)>(double)60)
							{
								privmsg(usr,NULL,"Minimal interval between searches is %d",conf.search_interval);
								usr->sent_search_interval=time(NULL);	
							}
							goto leave;
						}

					usr->last_search=time(NULL);					

					if (!strcasecmp(cmd,"Hub"))
					{

						//if passive mode is not allowed and user has no permission to be passive
						if (!conf.allow_passive && usr->perm && !strchr(usr->perm,'p'))
						{
							pubmsg(usr,"PASSIVE mode is not allowed.");
							pubmsg(usr,"Change your connection settings and reconnect.");
							debug(DEBUG_NET,"'%s' is kicked because PASSIVE mode is not allowed.",usr->nick);
							//kicking user from hub
							user_set_state(usr,STATE_QUIT);
							goto end;
						}

						if (!(par=strtok_r(NULL,sep,&__strtok_temp__))) goto end;

						if (strcasecmp(usr->nick,par))
						{
//							debug(DEBUG_PROTO,"On Search from %s@%s: Fake nick not allowed. (%s)",usr->nick,usr->ip,line);
							goto end;
						}

					}else
						{
							if (!conf.allow_forwarding && usr->perm && !strchr(usr->perm,'f') && strcmp(usr->ip,cmd))
							{
//								debug(DEBUG_PROTO,"On Search from %s@%s: Incorrect IP. (%s)",usr->nick,usr->ip,line);
								goto end;
							}
						}

					sendf(NULL,"%s|",line);

					goto end; /* }}} */

			/* 9. $MultiSearch {{{ */
			case 9:
					if(!conf.allow_search || !par || !user_tst_state(usr,STATE_REGISTERED)) goto end;
//					debug(DEBUG_PROTO,"On MultiSearch form %s@%s: Not supported yet. (%s)",usr->nick,usr->ip,line);
					goto end; /* }}} */

			/* 10. $To: {{{ */
			case 10:
					if ((i=nick2id(par))<0)
					{
						/* get rid of some client-side junk */
//						if (strcmp(par,"Vandel\\Debug")) debug(DEBUG_PROTO,"On PrivMsg from %s@%s: No such user. (%s)",usr->nick,usr->ip,line);
						goto end;
					}

					if (!(par=strtok_r(NULL,sep,&__strtok_temp__))) goto end;
					if (strcasecmp(par,"From:")) goto end;
					if (!(par=strtok_r(NULL,sep,&__strtok_temp__))) goto end;

					if (strcasecmp(usr->nick,par))
					{
//						debug(DEBUG_PROTO,"On PrivMsg from %s@%s: Fake nick not allowed. (%s)",usr->nick,usr->ip,line);
						goto end;
					}

					if (!(par=strtok_r(NULL," $",&__strtok_temp__))) goto end;
					if (!(par=strtok_r(NULL,"",&__strtok_temp__))) goto end;

					// if the user[n] is not OP or sOP, and usr has no permission to private chat then ontinue;
					if (user_tst_penalty(usr,PENALTY_NOPRVCHAT) && usr->perm && !strchr(user[i]->perm,'s') && !strchr(user[i]->perm,'s')) goto end;

					if (user[i]->cons) sendf(user[i],"[%s@%s] %s\r\n",usr->nick,usr->ip,par);
					else sendf(user[i],"%s|",line);

					goto end; /* }}} */

			/* 11. $Version {{{ */
			case 11:
					if (!par) goto end;
					if (!*par) goto end;
					if (strlen(par)+1>VER_LEN) goto end;

					my_duplicate(par,&usr->ver);
					goto end; /* }}} */

			/* 12. $ConnectToMe {{{ */
			case 12:
					if(!conf.allow_downloads || !par || !user_tst_state(usr,STATE_REGISTERED) || user_tst_penalty(usr,PENALTY_NODL)) goto end;

					if ((i=nick2id(par))<0)
					{
//						debug(DEBUG_PROTO,"On ConnectToMe from %s@%s: No such user. (%s)",usr->nick,usr->ip,line);
						goto end;
					}

					if (user[i]->cons)
					{
						privmsg(usr,NULL,"Can't connect to console user.");
						goto end;
					}

					if (!(par=strtok_r(NULL,":",&__strtok_temp__))) goto end;

					if (!conf.allow_forwarding && usr->perm && !strchr(usr->perm,'f') && strcmp(usr->ip,par))
					{
//						debug(DEBUG_PROTO,"On ConnectToMe from %s@%s: Incorrect IP. (%s)",usr->nick,usr->ip,line);
						goto end;
					}

					sendf(user[i],"%s|",line);
					goto end; /* }}} */

			/* 13. $MultiConnectToMe {{{ */
			case 13:
					if(!par || !user_tst_state(usr,STATE_REGISTERED)) goto end;

					if ((i=nick2id(par))<0)
					{
//						debug(DEBUG_PROTO,"On MultiConnectToMe from %s@%s: No such user. (%s)",usr->nick,usr->ip,line);
						goto end;
					}

					if (user[i]->cons)
					{
						privmsg(usr,NULL,"Can't connect to console user.");
						goto end;
					}

					if (!(par=strtok_r(NULL,":",&__strtok_temp__))) goto end;

					if (!conf.allow_forwarding && usr->perm && !strchr(usr->perm,'f') && strcmp(usr->ip,par))
					{
//						debug(DEBUG_PROTO,"On MultiConnectToMe from %s@%s: Incorrect IP. (%s)",usr->nick,usr->ip,line);
						goto end;
					}

					sendf(user[i],"%s|",line);
					goto end; /* }}} */

			/* 14. $RevConnectToMe {{{ */
			case 14:
					if(!conf.allow_downloads || !par || !user_tst_state(usr,STATE_REGISTERED) || user_tst_penalty(usr,PENALTY_NODL)) goto end;

					//if passive mode is not allowed and user has no permission to be passive
					if (!conf.allow_passive && !strchr(usr->perm,'p'))
					{
						pubmsg(usr,"PASSIVE mode is not allowed.");
						pubmsg(usr,"Change your connection settings and reconnect.");
						debug(DEBUG_NET,"'%s' is kicked because PASSIVE mode is not allowed.",usr->nick);
						//kicking user from hub
						user_set_state(usr,STATE_QUIT);
						goto end;
					}

					if (strcasecmp(usr->nick,par))
					{
						debug(DEBUG_PROTO,"On RevConnectToMe from %s@%s: Fake nick not allowed. (%s)",usr->nick,usr->ip,line);
						goto end;
					}

					if (!(par=strtok_r(NULL,sep,&__strtok_temp__))) goto end;

					if ((i=nick2id(par))<0)
					{
						debug(DEBUG_PROTO,"On RevConnectToMe from %s@%s: No such user. (%s)",usr->nick,usr->ip,line);
						goto end;
					}

					if (user[i]->cons)
					{
						privmsg(usr,NULL,"Can't connect to console user.");
						goto end;
					}

					sendf(user[i],"%s|",line);
					goto end; /* }}} */

			/* 15. $UniSearch {{{ */
			case 15:
					if(!conf.allow_search || !par || !user_tst_state(usr,STATE_REGISTERED)) goto end;

					debug(DEBUG_PROTO,"On UniSearch from %s@%s: Not supported yet. (%s)",usr->nick,usr->ip,line);
					privmsg(usr,NULL,"UniSearch not supported yet.");
					goto end; /* }}} */

			/* 16. $SR {{{ */
			case 16:
					if(!conf.allow_search || !par || !user_tst_state(usr,STATE_REGISTERED)|| user_tst_penalty(usr,PENALTY_NOSEARCH) ) goto end;

					if (strcasecmp(usr->nick,par))
					{
						debug(DEBUG_PROTO,"On SR from %s@%s: Fake nick not allowed. (%s)",usr->nick,usr->ip,line);
						goto end;
					}

					my_duplicate(line,&line_copy);

					par=strrchr(line_copy,'\005');
					if(!par) goto end;

					*par=0;
					par++;
					i=nick2id(par);
					if (i<0) goto end;

					sendf(user[i],"%s|",line_copy);
					goto end; /* }}} */

			/* 17. $OpForceMove {{{ */
			case 17:
					if(!par || !user_tst_state(usr,STATE_REGISTERED)) goto end;

					if (usr->perm && !strchr(usr->perm,'a') && !strchr(usr->perm,'s'))
					{
						debug(DEBUG_NET,"'%s'(not OP) tried to redirect someone",usr->nick);
						pubmsg(usr,"You are Not an Op");
						goto end;
					}

					my_duplicate(line,&line_copy);

					tmp1=strtok_r(line_copy," ",&__strtok_temp__); //$OpForceMove:
					tmp1=strtok_r(NULL,":",&__strtok_temp__); //$Who:
					tmp1=strtok_r(NULL,"$",&__strtok_temp__);//<victimNick>

					tmp2=strtok_r(NULL,":",&__strtok_temp__);//$Where:
					tmp2=strtok_r(NULL,"$",&__strtok_temp__);// <IP>

					tmp3=strtok_r(NULL,":",&__strtok_temp__)+4;// $Msg:<message>

					i=nick2id(tmp1);

					if (!strcmp(usr->nick,tmp1))
					{
						privmsg(usr,NULL,"You are trying to move yourself");
						goto end;
					}

					if (i<0)
					{
						pubmsg(usr,"You are trying to redirect person, that doesn't exist");
						goto end;
					}

					if (usr->perm!=NULL && user[i]->perm!=NULL)
					{
						if (strchr(user[i]->perm,'a') && strchr(usr->perm,'a'))
						{
							privmsg(usr,NULL,"You are trying to redirect OP");
							goto end;
						}

						if (strchr(user[i]->perm,'s') && (strchr(usr->perm,'a') || strchr(usr->perm,'s')))
						{
							privmsg(usr,NULL,"You are trying to redirect sOP");
							goto end;
						}

						if (usr->perm && strchr(user[i]->perm,'a') && strchr(usr->perm,'s')) privmsg(usr,NULL,"You are redirecting OP");
					}
					else goto end;

					privmsg(user[i],NULL,tmp3);

					sendf(user[i],"$ForceMove %s|",tmp2);
					user_set_state(user[i],STATE_QUIT);

					debug(DEBUG_NET,"'%s'(OP) moved '%s' because: '%s'",usr->nick,tmp1,tmp3);
					goto end; /* }}} */

			case 18:
					sendf(usr,"$Supports NoGetINFO NoHELLO|");
					goto end;

			default:
					debug(DEBUG_PROTO,"On %s from %s@%s: Unknown command. (%s)",cmd,usr->nick,usr->ip,line);
					goto leave;
		}
		/* }}} */
	end:
		my_free(line);
	}

leave:
		my_free(line);
		my_free(line_copy);

//		debug(DEBUG_PROTO,"manage_sock(%s) leave",usr->nick);
} /* }}} */

int is_managed(int No,int usr_no)
{
	int RC=0, i;

//	debug(DEBUG_PROTO,"is_managed(%d,%d) enter",No,usr_no);

	pthread_mutex_lock(&mutex_is_managed);

	if (usr_no>=0)
	{
		for (i=0;!RC && i<MAX_N_USER_MANAGER; i++)
			if (managed[i]==usr_no) RC=1;

		if (!RC && No>=0) managed[No]=usr_no;
		
	} else managed[No]=-1;
		
	pthread_mutex_unlock(&mutex_is_managed);
//	debug(DEBUG_PROTO,"is_managed(%d) leave",RC);
	return RC;
}


void test_slots( userrec_t *usr)
{	
	if (usr->slots>999 || usr->slots<0)
	{
		pubmsg(usr,"More than 999 slots is not allowed");
		return;
	}

	if(usr->slots>=conf.minslots)
	{
		user_set_state(usr,STATE_REGISTERED);
		dc_myinfo(NULL,usr);

		if(usr->slots==usr->slots_old) return;
						
		debug(DEBUG_NET,"'%s' minslots checked",usr->nick);
		pubmsg(usr,"The minimal amount of slots reached");
		pubmsg(usr,"-----------------------------------");
		
		if (conf.conf_motd && !usr->slots_old) 
		{	
			playfile(usr,conf.conf_motd);
			pubmsg(usr,"-----------------------------------");			
		}
		usr->slots_old=usr->slots;
		return;
	}

	pubmsg(usr,"The minimum number of upload slots is %d.",conf.minslots);

	if (conf.redirect_switch&REDIRECT_SWITCH_MINSLOTS)
	{
		pubmsg(usr,"You are being redirected to %s",conf.redirect_minslots);
		sendf(usr,"$ForceMove %s|",conf.redirect_minslots);
	}
	else pubmsg(usr,"Update its number and reconnect");

	debug(DEBUG_NET,"'%s' %s because of the amount of upload slots", usr->nick, conf.redirect_switch&REDIRECT_SWITCH_MINSLOTS?"redirected":"kicked");
	user_set_state(usr,STATE_QUIT);
}


void user_manager(void *voidp)
{
	int i,No;
	int_voidp_t parameter;
	struct timeval timeout;
	char *message=NULL;

	fd_set _sockfdset;
	fd_set *sockfdset=&_sockfdset;

	parameter.voidp=voidp;
	No=parameter.integer;

	debug(DEBUG_PROTO,"user_manager(%d)",No);

	FD_ZERO(sockfdset);

	while(1)
	{

	/* we use 1ms timeout because we want select() to proces new users even when
	 * there is no traffic on the sockets */
		usleep(100);
		if (conf.minimal_sleep_time<0) usleep(-conf.minimal_sleep_time);		

		pthread_mutex_lock( &mutex_user_manager[No]);

       	timeout.tv_sec=0;
		timeout.tv_usec=1000;

		for(i=0;i<NUSERS;i++) if (user[i]) FD_SET(user[i]->sock,sockfdset);
		select(maxfd,sockfdset,NULL,NULL,&timeout);

		/* this needs to be a reverse loop because of the method used in deluser() */
		for(i=0;i<NUSERS;i++)
		{

			if (conf.minimal_sleep_time>MAXUSERS) usleep(conf.minimal_sleep_time/(int)(NUSERS+1));

			if (user[i]  && !user_tst_state(user[i],STATE_QUIT) && !is_managed(No,i))
			{
				if (conf.minslots && user_tst_state(user[i],STATE_SLOTS_TESTED)) test_slots(user[i]);

				if (user_tst_state(user[i],STATE_WELCOME))
				{
				
				    if (*(conf.conf_welcome)) 
				    {
						playfile(user[i],conf.conf_welcome);
						pubmsg(user[i],"-----------------------------------");
				    }

    				    /* host is banned permanently*/
				    if (validhost(user[i],conf.conf_banned,1)) user_tst_state(user[i],STATE_QUIT);

				    user_set_state(user[i],STATE_CONNECTED);
				}

				if (FD_ISSET(user[i]->sock,sockfdset) && my_recv(user[i],&message))
				{
					 msg2buff(user[i],message);
					 my_free(message);
				}

				if (user[i]->buf) manage_sock(user[i]);
			    
				is_managed(No,-1);
			}
		}
		pthread_mutex_unlock( &mutex_user_manager[No] );
	}

	debug(DEBUG_WRNG,"manager(): this should be never reached!");
	exit(1); /* never reached */
}


/* init_var() - install the signal handler and initialize variables {{{ */
void init_var(char *argv0)
{
	struct passwd *user_data;
	struct group *group_data;

	/* clear main variables */
	NUSERS=0;
	NPENAL=0;
	DAEMON=0;

	/* install signal handler */

	signal(SIGPIPE,finish);
	signal(SIGHUP,finish);
	signal(SIGINT,finish);
	signal(SIGQUIT,finish);
	signal(SIGTERM,finish);
	signal(SIGSEGV,finish);
	signal(SIGKILL,finish);

	memset(penalties,0,MAXPENALTIES*sizeof(Tpenalty *));
	memset(user,0,MAXUSERS*sizeof(userrec_t *));
	memset(&conf,0,sizeof(config_t));

	/* set up default config values */

	conf.conf_main=(char *)my_malloc(strlen(CONFIGDIR)+strlen(CONF_MAIN)+2); /* one byte for '/' and one for '\0' */
	sprintf(conf.conf_main, "%s%s%s", CONFIGDIR, (CONFIGDIR[strlen(CONFIGDIR)-1]=='/')?"":"/", CONF_MAIN);

	conf.conf_welcome=strdup(CONF_WELCOME);
	conf.conf_motd=strdup(CONF_MOTD);
	conf.conf_banned=strdup(CONF_BANNED);
	conf.conf_cusers=strdup(CONF_CUSERS);
	conf.conf_callow=strdup(CONF_CALLOW);
	conf.conf_nallow=strdup(CONF_NALLOW);
	conf.conf_penalties=strdup(CONF_PENALTIES);
	conf.conf_rules=strdup(CONF_RULES);

	conf.log_level=LOG_LEVEL;
	conf.log_dir=strdup(LOG_DIR);
	conf.log_main=strdup(LOG_MAIN);
	conf.log_net=strdup(LOG_NET);
	conf.log_proto=strdup(LOG_PROTO);
	conf.log_wrng=strdup(LOG_WRNG);
	conf.log_con=strdup(LOG_CON);
	conf.log_chat=strdup(LOG_CHAT);
	conf.listen_interface=strdup(LISTEN_INTERFACE);
	conf.admin_contact=strdup(ADMIN_CONTACT);
	conf.listen_port_main=LISTEN_PORT_MAIN;
	conf.listen_port_cons=LISTEN_PORT_CONS;

	conf.redirect_switch=REDIRECT_SWITCH_OFF;
	conf.redirect_access=strdup(REDIRECT_ACCESS);
	conf.redirect_hub_is_full=strdup(REDIRECT_HUB_IS_FULL);
	conf.redirect_minshare=strdup(REDIRECT_MINSHARE);
	conf.redirect_minslots=strdup(REDIRECT_MINSLOTS);

	conf.penalties_update_interval=PENALTIES_UPDATE_INTERVAL;
	conf.search_interval=SEARCH_INTERVAL;

	conf.allow_downloads=ALLOW_DOWNLOADS;
	conf.allow_search=ALLOW_SEARCH;

	conf.n_listen_main=0;
	conf.max_commands=MAX_COMMANDS;

	user_data=getpwnam(DEFAULT_USER);
	if(!user_data)
	{
		debug(DEBUG_STD,"User `%s' not found - terminating.",DEFAULT_USER);
		exit(0);
	}

	conf.uid=user_data->pw_uid;
	conf.user=strdup(DEFAULT_USER);


	group_data=getgrnam(DEFAULT_GROUP);

	if(!group_data)
	{
		debug(DEBUG_STD,"Group `%s' not found - terminating.",DEFAULT_GROUP);
		exit(0);
	}
	
	conf.group=strdup(DEFAULT_GROUP);

	conf.gid=group_data->gr_gid;

	conf.hubname=strdup(HUBNAME);

	conf.userlimit=USERLIMIT;
	conf.allow_broken_key=ALLOW_BROKEN_KEY;
	conf.allow_chat=ALLOW_CHAT;
	conf.allow_passive=ALLOW_PASSIVE;
	conf.allow_passive=ALLOW_FORWARDING;
	conf.minslots=0;
	conf.minshare=0;

	conf.std_penalty_duration=STD_PENALTY_DURATION;

	conf.idle_timeout=PING_TIMEOUT;
	conf.register_timeout=REGISTER_TIMEOUT;

	conf.minimal_sleep_time=MINIMAL_SLEEP_TIME;

	conf.max_chat_length=MAX_CHAT_LENGTH;
	conf.kick_max_chat_length=KICK_MAX_CHAT_LENGTH;
	/* end of default config values */

	/* logging file handlers */
	file_net=NULL;
	file_proto=NULL;
	file_wrng=NULL;
	file_chat=NULL;

	/* tcpd syslog messages */
	allow_severity=LOG_INFO;
	deny_severity=LOG_WARNING;


	/* socket options*/
	listen_opts.so_debug=0;
	listen_opts.so_broadcast=0;
	listen_opts.so_reuseaddr=1;
	listen_opts.so_keepalive=1;

	listen_opts.so_linger.l_onoff=1;
	listen_opts.so_linger.l_linger=0;

	listen_opts.so_oobinline=1;
	listen_opts.so_sndbuf=VAL_SO_SNDBUF;
	listen_opts.so_rcvbuf=VAL_SO_RCVBUF;
	listen_opts.so_dontroute=0;
	listen_opts.so_rcvlowat=1;
	listen_opts.so_rcvtimeo=0;
	listen_opts.so_sndlowat=1;
	listen_opts.so_sndtimeo=0;

	user_opts.so_debug=0;
	user_opts.so_broadcast=0;
	user_opts.so_reuseaddr=0;
	user_opts.so_keepalive=1;

	user_opts.so_linger.l_onoff=1;
	user_opts.so_linger.l_linger=0;

	user_opts.so_oobinline=1;
	user_opts.so_sndbuf=VAL_SO_SNDBUF;
	user_opts.so_rcvbuf=VAL_SO_RCVBUF;
	user_opts.so_dontroute=0;
	user_opts.so_rcvlowat=1;
	user_opts.so_rcvtimeo=0;
	user_opts.so_sndlowat=1;
	user_opts.so_sndtimeo=0;

	/* other */
	__progname=get_progname(argv0);
	randomize();
	time(&starttime);
} /* }}} */


void cleaner()
{
	int i;

	time_t time_;

	/* main loop - finish only when error or killed with a signal */
	while(1)
	{
		sleep(CLEANER_INTERVAL);
		lock_threads();
//		debug(DEBUG_PROTO,"cleaner_in");

		time(&time_);
		if (conf.penalties_update_interval && time_-last_penalties_update>conf.penalties_update_interval)
		{
			last_penalties_update=time_;
			penalties_write();
		}

		for(i=0;i<NUSERS;i++)
			if (user[i])
			{
				check_timeout(user[i]);
				if (user_tst_state(user[i],STATE_QUIT)) deluser(user[i],NULL);
			}

		unlock_threads();
//		debug(DEBUG_PROTO,"cleaner_out");
	}
}

/* main() - main program {{{ */
int main(int argc,char *argv[])
{
	int_voidp_t i;

	/* init variables */
	init_var(argv[0]);
	maxfd=getdtablesize();

	if(maxfd>FD_SETSIZE) maxfd=FD_SETSIZE;

	/* parse command-line options */
	parse_options(&conf,argc,argv);

	/* read config file */
	if (parse_config(1,&conf)) exit(0);

	make_sockets();

	/* set uid/gid to proper user/group */
	setuid(conf.uid);
	setgid(conf.gid);

	/* compose direct paths to files for further use */
	make_paths(&conf);

	if (conf.userlimit>MAXUSERS)
	{
		debug(DEBUG_STD,"User limit (%d) is higher than MAXUSERS (%d). See main.h and recompile.",conf.userlimit,MAXUSERS);
		exit(0);
	}

	/* daemonize */
	if (daemon(0,0)<0) debug(DEBUG_STD,"Couldn't fall into background: %s.", strerror(errno));
	else
		{
			DAEMON=1;

			if ((*(conf.log_main))&&(conf.log_level&DEBUG_STD))
			if (freopen(conf.log_main,"a",stderr)==NULL)
				debug(DEBUG_SYS,"Couldn't open main log file (%s): %s.", conf.log_main, strerror(errno));

			debug(DEBUG_SYS|DEBUG_STD,"Daemon started.");
		}

		/* open rest of the log files */

	if ((*(conf.log_net)) && (conf.log_level&DEBUG_NET) && (file_net=fopen(conf.log_net,"a"))==NULL) debug(DEBUG_SYS|DEBUG_STD,"Couldn't open network access log file (%s): %s.",conf.log_net,strerror(errno));

	if ((*(conf.log_proto)) && (conf.log_level&DEBUG_PROTO) && (file_proto=fopen(conf.log_proto,"a"))==NULL) debug(DEBUG_SYS|DEBUG_STD,"Couldn't open protocol log file (%s): %s.",conf.log_proto,strerror(errno));

	if ((*(conf.log_wrng)) && (conf.log_level&DEBUG_WRNG) && (file_wrng=fopen(conf.log_wrng,"a"))==NULL) debug(DEBUG_SYS|DEBUG_STD,"Couldn't open warnings log file (%s): %s.",conf.log_wrng,strerror(errno));

	if ((*(conf.log_con)) && (conf.log_level&DEBUG_CON) && (file_con=fopen(conf.log_con,"a"))==NULL) debug(DEBUG_SYS|DEBUG_STD,"Couldn't open console log file (%s): %s.",conf.log_con,strerror(errno));

	if ((*(conf.log_chat)) && (conf.log_level&DEBUG_CHAT) && (file_chat=fopen(conf.log_chat,"a"))==NULL) debug(DEBUG_SYS|DEBUG_STD,"Couldn't open chat log file (%s): %s.",conf.log_chat,strerror(errno));

	/* initialize semaphores */
	if (pthread_mutex_init(&mutex, 0) ||
		pthread_mutex_init(&mutex_penalties, 0) ||
		pthread_mutex_init(&mutex_is_managed, 0) ||
		pthread_mutex_init(&mutex_inet_ntoa, 0)||
		pthread_mutex_init(&mutex_my_hosts_access, 0)||
		pthread_mutex_init(&mutex_listen_sr,0)||
		pthread_mutex_init(&mutex_listen_cons,0)||
		pthread_mutex_init(&mutex_th_cleaner,0)||
		pthread_mutex_init(&mutex_my_send,0) ||
		pthread_mutex_init(&mutex_myinfo,0))
	{
		debug(DEBUG_SYS|DEBUG_STD,"Couldn't create synchronization mutex: %s - terminating.", strerror(errno));
		exit(1);
	}

	penalties_read();

	for(i.integer=0;i.integer<MAX_N_LISTEN_MAIN;i.integer++)
	{
		if (pthread_mutex_init(&mutex_listen_main[i.integer],0))
		{
			debug(DEBUG_SYS|DEBUG_STD,"Couldn't create synchronization mutex: %s - terminating.", strerror(errno));
			exit(1);
		}

		/* create listening thread for main service */
		if (i.integer<conf.n_listen_main &&
			(listen_main==-1 || pthread_create( &listen_th_main[i.integer], NULL, (void *)listen_thread_main, i.voidp) ))
		{
			debug(DEBUG_SYS|DEBUG_STD,"Couldn't create the main service thread - terminating: %s.", strerror(errno));
			exit(1);
		}
	}

	for(i.integer=0;i.integer<MAX_N_USER_MANAGER;i.integer++)
	{
		managed[i.integer]=-1;

		if (pthread_mutex_init(&mutex_user_manager[i.integer],0))
		{
			debug(DEBUG_SYS|DEBUG_STD,"Couldn't create synchronization mutex: %s - terminating.", strerror(errno));
			exit(1);
		}

		/* create listening thread for main service */
		if (i.integer<conf.n_user_manager &&
			(pthread_create( &th_user_manager[i.integer], NULL, (void *)user_manager, i.voidp) ))
		{
			debug(DEBUG_SYS|DEBUG_STD,"Couldn't create the user manager thread - terminating: %s.", strerror(errno));
			exit(1);
		}
	}

	if ((pthread_create( &th_cleaner, NULL, (void *)cleaner, NULL)!=0))
	{
		debug(DEBUG_SYS|DEBUG_STD,"Couldn't create Cleaner thread - terminating: %s.", strerror(errno));
		exit(1);
	}

	if (conf.minslots && (listen_sr==-1 || pthread_create( &listen_th_sr, NULL, (void *)listen_thread_SR, NULL)!=0))
	{
		debug(DEBUG_SYS|DEBUG_STD,"Couldn't create the testing minslots thread - terminating: %s.", strerror(errno));
		exit(1);
	}

	/* create listening thread for remote console */
	if (conf.listen_port_cons && (listen_cons==-1 || pthread_create( &listen_th_cons, NULL, (void *)listen_thread_cons, NULL)!=0))
	{
		debug(DEBUG_SYS|DEBUG_STD,"Couldn't create the remote console thread - console disabled: %s.", strerror(errno));
		exit(1);
	}

	while(1)
	{
		sleep(1);
		if (rehash)
		{
			finish(SIGHUP);
			rehash=0;
		}
	}

	debug(DEBUG_WRNG,"main(): this should be never reached!");
	exit(1); /* never reached */
	return 0;
} /* }}} */

/* VIM Settings {{{
* Local variables:
* tab-width: 14
* c-basic-offset: 4
* soft-stop-width: 4
* c indent on
* End:
* vim600: sw=4 ts=4 sts=4 cindent fdm=marker
* vim<600: sw=4 ts=4
* }}} */
