/* Intro {{{
 * ----------------------------------------------------------------
 * DConnect Daemon
 *
 *	 #(@) Copyright (c) 2002, DConnect development team
 *	 #(@) Homepage: http://www.dc.ds.pg.gda.pl/
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
 * MA 02139, USA.
 *
 * ----------------------------------------------------------------
 * }}} */

 /* include local header files {{{ */
#include "pch.h"
#include "main.h"
#include "match.h"
#include "misc.h"
#include "net.h"
#include "penalties.h"
#include "cmd.user.h"
#include "conf.h"

/* }}} */

extern int NPENAL; /* penalties count */
extern int NUSERS; /* penalties count */

extern Tpenalty *penalties[MAXPENALTIES]; /* table of penalties */
extern userrec_t *user[MAXUSERS];	/* users table */

extern pthread_mutex_t mutex_penalties;	/* provides synchronization between threads */
extern int test_penalty_counter;

/*
penalties_find_id( crit, *str )

.description:
	gets the penalties id.

.args
	int crit		- IN - SRCH_IP
							SRCH_NICK
							SRCH_BOTH

	char *str		- IN - the searching string

.return values:
	-1		- the ip is not present on the penalties list
	value	- the number of the ithem containing the >ip< is >value<
{{{*/

int penalties_find_id(int crit, char *str)
{
	int i;

//	debug(DEBUG_PROTO,"penalty_find_id() enter");


	for(i=0;i<NPENAL;i++)
		if (penalties[i] &&
			((crit>=0 && match(penalties[i]->ip, str)) ||
			 (crit<=0 && match(penalties[i]->nick, str)))) goto leave;
						 

//	debug(DEBUG_PROTO,"penalty_find_id(-1) leave");
	return -1;
	
leave:
//	debug(DEBUG_PROTO,"penalty_find_id(%d) leave",i);
	return i;	
} /* }}} */

/*
void user_penalty(*usr)

.description:
 checks if the user->ip is listed on >penalties< list and changes user->penalty
 to fit those from the list

.args:
	userrec_t *usr - IN(ip) & OUT(penalty) - userrec
{{{*/

void user_penalty(userrec_t *usr)
{
	int nr;//itemnumber

	nr=penalties_find_id(SRCH_IP,usr->ip);

	if (nr<0)
	{
		usr->penalty=NULL;
		return;
	}

	usr->penalty=penalties[nr];

} /* }}} */

/*
void addpenalty(*penrec)

.description:
 adds the peanlty to penalties list and userrecs

.args:
	Tpenalty *penrec - penalty record

{{{*/

void addpenalty(Tpenalty *penrec)
{
//	debug(DEBUG_PROTO,"addpenalty() enter");

	/* FIXME: find penalty wich ends in shortest amount of time */
	if (NPENAL+1>MAXPENALTIES)
	{
		my_free(penrec->ip);
		my_free(penrec->reason);
		my_free(penrec->nick);
		my_free(penrec->op);
		my_free(penrec);
		return;
	}

	penalties[NPENAL++]=penrec; /* change the item when it exists */
	penalmsg(penrec);	/* display info about changed penalty; kicks baned */

//	debug(DEBUG_PROTO,"addpenalty() leave");
}/* }}} */

/*
void delpenalty(no, block)

.description:
 removes the peanlty to penalties list and userrecs by itemnuber in penalties[]

.args:
	int no - itemnumber of user's penalty in penalties list
	int block - use mutexes while deleting??

{{{*/

void delpenalty( int no )
{
	int i;
	Tpenalty *penrec=penalties[no];

//	debug(DEBUG_PROTO,"delpenalty() enter");
	if (!penrec) return;

	/* remove penalty from userrecs */
	for (i=0;i<NUSERS;i++)
		if (user[i] && user[i]->penalty==penrec) user[i]->penalty = NULL;

	/* remove penalty from penalties[] */

	penalties[no]=penalties[--NPENAL];
	penalties[NPENAL]=NULL;

	/* free penrec */
	my_free(penrec->ip);
	my_free(penrec->reason);
	my_free(penrec->nick);
	my_free(penrec->op);
	my_free(penrec);

//	debug(DEBUG_PROTO,"delpenalty() leave");
}/* }}} */

/*
user_set_penalty()
.description:
	sets user's penalty

.args
	char *ip		- user's ip
	char *nick		- user's nick(only PENALTY_NONE)
	char *op		- punisher
	int penalty		- the penalty {0-none}
	unsigned long	start_date - START_DATE
	long duration	- duration of the penalty {in minutes?)
	char *reason	- reason of penalty

{{{*/

void user_set_penalty(char *ip,char *nick, char *op,int penalty, unsigned long start_date,long duration, char *rsn, int only_memory)
{
	int nr;		/* penalty list itemnumber */
	Tpenalty *penrec=NULL; /* penalty record (used to add new penalty) */
	char *reason=NULL;

	pthread_mutex_lock(&mutex_penalties);
//	debug(DEBUG_PROTO,"user_set_penalty('%s','%s',%d) enter",ip, nick,penalty);

	/* removing penalty (when penalty==PENALTY_NONE) {{{ */
	if (penalty==PENALTY_NONE)
	{

		nr=penalties_find_id(SRCH_IP,ip);
		if (nr<0) nr=penalties_find_id(SRCH_NICK,nick);

		//if found delete the item conataining the user's ip, from the penalties list
//		if (nr>=0) delpenalty(nr,1);

		if (nr>=0) penalties[nr]->end_date=penalties[nr]->start_date;
		goto leave;
	} /* }}} */

	/* if no reason */
	if (!rsn) my_duplicate("NO REASON",&reason);
	else my_duplicate(rsn,&reason);

	nr=penalties_find_id(SRCH_IP,ip); //punishments by ip ONLY!

// if (nr<0) nr=penalties_find_id(SRCH_NICK,nick);
//	pubmsg(NULL,"IP=%s, NICK=%s, NR=%d",ip,nick,nr);

		if (nr<0) /* if penalety does not yet exist */
		{
				penrec=(Tpenalty *) my_malloc(sizeof(Tpenalty));		/* create a penalty record... */
				memset(penrec,0,sizeof(Tpenalty));

				my_duplicate(ip,&penrec->ip);
				my_duplicate(nick,&penrec->nick);
				my_duplicate(op,&penrec->op);

				penrec->penalty=penalty;
				penrec->reason=reason;

				penrec->start_date=start_date;
				penrec->end_date=(unsigned long)duration+penrec->start_date;

				addpenalty(penrec); /* ...to add it */
				goto leave;
		}

	penalties[nr]->penalty = penalty;	/* penalty overwrites existing */

	my_duplicate(reason,&penalties[nr]->reason);	/* set the new reason of penalty */
	my_free(reason);

	my_duplicate(nick, &penalties[nr]->nick);	/* set the new nick */
	my_duplicate(op, &penalties[nr]->op);

	penalties[nr]->start_date = start_date;
	penalties[nr]->end_date = (unsigned long)duration+penalties[nr]->start_date;

	penalmsg(penalties[nr]);	/* display info about changed penalty; kicks baned */

leave:
	if (!only_memory) penalties_write();
//	debug(DEBUG_PROTO,"user_set_penalty() leave");

	pthread_mutex_unlock(&mutex_penalties);
} /* }}} */

/*
user_exp_penalty(penalty)

.description:
	tests if penalty expired

.args
	Tpenalty *penalty

.return values:
	<=0: the >penalty< expired
	>0: the >penalty< duration
{{{*/

long user_exp_penalty(Tpenalty *penalty)
{
//	debug(DEBUG_PROTO,"%ld %lu %lu %lu",sizeof(time_t),penalty->start_date,penalty->end_date,(unsigned long)time(NULL));

	return (long)((time_t)penalty->end_date-time(NULL));
} /* }}} */

/*
user_tst_penalty(usr, penalty, duration)

.description:
	tests user's for penalties

.args
	userrec_t *usr - IN - userrec
	int penalty			- IN - the penalty (0 - none)

.return values:
	0: the >penalty< does NOT match user's penalty, the user has no penalty;
	1: the >penalty< DOES match user's penalty
{{{*/

int user_tst_penalty(userrec_t *usr, int penalty)
{
	int code=0;
/*	fd_set fdset; */

//	debug(DEBUG_PROTO,"usr_tst_penalty() enter");

	if (usr->penalty)
	{
		code = (usr->penalty->penalty&penalty);

		// if penalty expired
		if (user_exp_penalty(usr->penalty)<=0)
		{
			//remove user's penalty
//			user_set_penalty(&usr->ip,usr->nick,NULL,PENALTY_NONE,0,0,NULL,0);
			code = 0;
		}
	}

//	debug(DEBUG_PROTO,"usr_tst_penalty() leave");
	return code;
} /* }}} */

